##
## Copyright (c) 2005-2007 by Rafael D'Halleweyn
##
## Licensed under Version 2 of the GPL
##

package ClusTerm::App;
use strict;
use warnings;
use Glib qw(TRUE FALSE);
use Gtk2;
use Gnome2;
use ClusTerm::Term qw(:roles);
use ClusTerm::Menus;
use ClusTerm::Preferences;
use ClusTerm::Utils;
use ClusTerm::Diff;
use Time::HiRes qw(time alarm);
use File::Basename;
use File::Spec::Functions;

# register with superclass of Gnome2::App
use ClusTerm::GlibSubclass Gnome2::App::;

our $title = "ClusTerm";
our $name = "clusterm";
our $VERSION = "0.6";
our $copyright = "Copyright  2005-2007 Rafael D'Halleweyn";
our $author = "Rafael D'Halleweyn";
our $about = "Cluster SSH Terminal";

our $icon;

sub new {
	my ($class, $clients, $options) = @_;
	my $self = Glib::Object::new(__PACKAGE__, $title);
	$self->{term} = [ ];

	# preferences
	$self->_create_prefs;
	# make sure that layout is valid
	$self->{prefs}->set('layout', $self->{prefs}->get('layout'));

	map { $self->{$_} = $options->{$_} } keys %$options;
	$self->{type} ||= $self->{prefs}->get('layout');
	@{$self->{qw(cols rows)}} = (1, 1);
	$self->{spacing} = [ 4, 4 ];
	unless ($self->_create_box(scalar @$clients)) {
		print STDERR "Unknown layout type \"$self->{type}\"\n";
		return undef;
	}

	# menus
	$self->_create_menus;

	$self->set_contents($self->{box});
	_get_icon($0) unless $icon;
	$self->set_icon_list($icon) if $icon;

	# need to show, otherwise we measure the wrong sizes
	# show_all is better?
	$self->show_all;
	#$self->set_my_geometry;

	# vtes
	for(my $i = 0; $i < @$clients; $i++) {
		$self->add_host($clients->[$i],
			$i ? ROLE_ACTOR : ROLE_DIRECTOR);
	}

	$self->{director} = $self->{term}[0];
	$self->{prev_director} = undef;

	$self->set_my_geometry;
	$self->resize_from_cols_rows(80, 24, TRUE);

	$self->{focused} = undef;
	$self->{term}[0]->grab_focus if (@{$self->{term}});
	$self->{zoom} = 1;

	$self->_update_menus;
	return $self;
}

sub _get_icon {
	my ($exe) = @_;

	my $theme = Gtk2::IconTheme->get_default();

	my $icon_path;
	my $theme_icon;
	if ($theme_icon = $theme->lookup_icon('clusterm', 64, [])) {
		$icon_path = $theme_icon->get_filename();
	} else {
		($icon_path = $0) =~ s{[^/]+$}{clusterm.png};
	}
	eval { $icon = Gtk2::Gdk::Pixbuf->new_from_file($icon_path) };
	print STDERR "Did not find icon clusterm.png\n" unless $icon;
}

sub _create_menus {
	my $self = shift;

	my $accel;
	my $menus = [ values %ClusTerm::Menus::menu ];
	&ClusTerm::Utils::uiinfo_accelerators($menus);
	&ClusTerm::Utils::uiinfo_swapped($menus, $self);
	&ClusTerm::Utils::uiinfo_icons($menus, $self);

	my $show_menubar = $self->{prefs}->get('show_menubar');
	# window menu-bar
	($self->{menubar}, $accel)
		= &ClusTerm::Menus::create_menu_bar;
	$self->add_accel_group($accel);
	$self->set_menus($self->{menubar}) if $show_menubar;

	# pre-create context menu, add it's AccelGroup to the window
	($self->{context_menu}, $accel)
		= &ClusTerm::Menus::create_context_menu;
	$self->add_accel_group($accel);

	# add shortcuts AccelGroup to the window
	$accel = &ClusTerm::Menus::create_shortcut_accel;
	$self->add_accel_group($accel);

	map { $_->set_active($show_menubar) } 
		@{$ClusTerm::Menus::menu{show_menubar}{widgets}};

	# set correct radio-item
	for my $item (@{$ClusTerm::Menus::menu{view_as}{moreinfo}}) {
		next unless $item->{value} eq $self->{type};
		$item->{widgets}[0]->set_active(1);
	}
}

sub _update_menus {
	my ($self) = @_;
	my $terms = $self->{term} ? scalar @{$self->{term}} : 0;

	for my $menu (values %ClusTerm::Menus::menu) {
		next unless exists $menu->{sensitive_when_terms};

		my $enable = $terms >= $menu->{sensitive_when_terms};
		map { $_->set_sensitive($enable) } @{$menu->{widgets}};
	}
}

sub _create_prefs {
	my $self = shift;

	# preferences is both the dialog and collection of the settings
	$self->{prefs} = ClusTerm::Preferences->new($self, $icon);
	$self->{prefs}->signal_connect_swapped(
		delete_event => sub { $_[0]->{prefs}->hide; TRUE; }, $self);
	$self->{prefs}->signal_connect_swapped(
		response => sub { $_[0]->{prefs}->hide; TRUE; }, $self);
	$self->{prefs}->signal_connect_swapped(
		notify => \&_prefs_change, $self);

	$self->{show_menus} = $self->{prefs}->get('show_menubar');
}

sub _prefs_change {
	my ($self, $prop, $prefs) = @_;
	return unless $prop->get_name eq 'profile';
	return unless @{$self->{term}};

	my $profile = $self->{prefs}->get('profile');

	my @size = (
		$self->{term}[0]->get_column_count,
		$self->{term}[0]->get_row_count
	);
	my $changes = 0;
	for (@{$self->{term}}) {
		next if $_->get('profile') eq $profile;
		$_->set('profile', $profile);
		$changes++;
	}

	return unless $changes;
	$self->set_my_geometry;
	$self->resize_from_cols_rows(@size);
}

sub set_my_geometry {
	my $self = shift;

	my $menu_height = $self->{show_menus}
		? $self->{menubar}->size_request->height : 0;

	if ($self->{term}[0]) {
		my @win_size;
		unless ($self->{win_size}) {
			my $app_alloc = $self->allocation;
			@win_size = ( $app_alloc->width,
				$app_alloc->height );
			@win_size = $self->size_request
				unless $win_size[0] && $win_size[1];
			$self->{win_size} = \@win_size;
		} else {
			@win_size = @{$self->{win_size}};
		}
		my @char_size = ($self->{term}[0]->get_char_width,
			$self->{term}[0]->get_char_height);
		my @term_pad = $self->{term}[0]->get_padding;
		my @win_pad = (0, $menu_height);
		if ($self->{type} =~ /^note/) {
			$win_pad[0] += $self->{notebook_tab_width};
			$win_pad[1] += $self->{notebook_tab_height};
		}
 
		my @win_base = $self->_pixel_size_from_term_cols_rows(0, 0,
			@char_size, @win_pad, @term_pad);
		# minimum characters that need to be visible to provide
		# sufficient space to the rest of the GUI
		my @min_chars = $self->_term_cols_rows_from_pixel_size(
			@win_size, @char_size, @win_pad, @term_pad, TRUE);
		$min_chars[0] = 20 if $min_chars[0] < 20;
		$min_chars[1] = 6 if $min_chars[1] < 6;

		my @win_min = $self->_pixel_size_from_term_cols_rows(
			@min_chars, @char_size, @win_pad, @term_pad);

		# the sizes are the sizes of the geometry widget!!!
		my $geom = Gtk2::Gdk::Geometry->new;
#print "Geometery base $win_base[0] x $win_base[1], min $win_min[0] x $win_min[1], char $char_size[0] x $char_size[1]\n";
#print "\tfrom ", (caller 1)[3], "\n";
		$geom->base_width($win_base[0]);
		$geom->base_height($win_base[1]);
		$geom->min_width($win_min[0]);
		$geom->min_height($win_min[1]);
		$geom->width_inc($char_size[0] * $self->{cols});
		$geom->height_inc($char_size[1] * $self->{rows});
		$self->set_geometry_hints($self, $geom,
			[ 'base-size', 'min-size', 'resize-inc' ]);

		$self->set_resizable(TRUE);
	} else {
		$self->set_resizable(FALSE);
	}
}

sub resize_from_cols_rows {
	my ($self, $c, $r, $limited) = @_;

	unless ($self->{term}[0]) {
		unless ($ClusTerm::Menus::menu{show_menubar}{widgets}
				[0]->get_active) {
			# have to show menubar, otherwise there is nothing!
			$self->{menubar}->show;
			map { $_->set_active(1) }
				@{$ClusTerm::Menus::menu{show_menubar}
					{widgets}};
		}
		my $req = $self->{menubar}->size_request;
		$self->resize($req->width, $req->height);
		return;
	}

	my $menu_height = $self->{prefs}->get('show_menubar')
		? $self->{menubar}->size_request->height : 0;
	my @win_pad = (0, $menu_height);
	if ($self->{type} =~ /^note/) {
		$win_pad[0] += $self->{notebook_tab_width};
		$win_pad[1] += $self->{notebook_tab_height};
	}
	my @char_size = ($self->{term}[0]->get_char_width,
		$self->{term}[0]->get_char_height);
	my @term_pad = $self->{term}[0]->get_padding;
#print "Term pad $term_pad[0] x $term_pad[1]\n";

	my @size = $self->_pixel_size_from_term_cols_rows(
		$c, $r, @char_size, @win_pad, @term_pad);

	if ($limited) {
		# but not bigger than the screen
		my $screen = $self->get_screen;
		my @max_chars = $self->_term_cols_rows_from_pixel_size(
			$screen->get_width * .9, $screen->get_height * .9,
			@char_size, @win_pad, @term_pad);
		my @win_max = $self->_pixel_size_from_term_cols_rows(
			@max_chars, @char_size, @win_pad, @term_pad);
		$size[0] = $win_max[0]
			if ($size[0] > $win_max[0]);
		$size[1] = $win_max[1]
			if ($size[1] > $win_max[1]);
	}

#print "Setting size to $size[0] x $size[1]\n";
#print "\tfrom ", (caller 1)[3], "\n";
	$self->resize(@size);
	# delayed resize, to make sure it happens
	my $count = 0;
	Glib::Timeout->add(100, \&_delayed_resize, [ $self, \$count, @size ]);
}

sub _delayed_resize {
	my ($self, $count, @size) = @{$_[0]};

	# sometimes we need to try harder to get the size we want, why?
	$$count++;
	if ($$count > 10) {
		print "Failed to resize window to ", join("x", @size), "\n";
		return 0;
	}

	my @now = $self->get_size();
#print "Size now $now[0] x $now[1]\n";
	if ($size[0] == $now[0] && $size[1] == $now[1]) {
		if ($$count > 0) {
		$self->resize($size[0], $size[1]);
			$$count = -1;
			return 1;
		} else {
			return 0;
		}
	}

	$self->resize(@size);
	return 1;
}

sub _pixel_size_from_term_cols_rows {
	my ($self, $c, $r);
	my (@char_size, @win_pad, @term_pad);
	($self, $c, $r, @char_size[0..1], @win_pad[0..1], @term_pad[0..1])
		= @_;

	return (
		($char_size[0] * $c + $term_pad[0]) * $self->{cols}
			+ $self->{spacing}[0] * ($self->{cols} - 1)
			+ $win_pad[0],
		($char_size[1] * $r + $term_pad[1]) * $self->{rows}
			+ $self->{spacing}[1] * ($self->{rows} - 1)
			+ $win_pad[1],
	);
}

sub _term_cols_rows_from_pixel_size {
	my ($self, $w, $h);
	my (@char_size, @win_pad, @term_pad, $round_up);
	($self, $w, $h, @char_size[0..1], @win_pad[0..1], @term_pad[0..1],
			$round_up)
		= @_;

	# size for one ClusTerm::Term
	my $c = $self->{cols} || 1;
	my $r = $self->{rows} || 1;
	$w = ($w - $win_pad[0] - $self->{spacing}[0] * ($self->{cols} - 1))
		/ $c;
	$h = ($h - $win_pad[1] - $self->{spacing}[1] * ($self->{rows} - 1))
		/ $r;

	my @round = (0, 0);
	@round = ($char_size[0] - 1, $char_size[1] - 1) if $round_up;

	return (
		int(($w - $term_pad[0] + $round[0]) / $char_size[0]),
		int(($h - $term_pad[1] + $round[1]) / $char_size[1]),
	);
}

sub context_menu {
	my ($self, $button, $obj, $term) = @_;

	$term->set("showing_context_menu", TRUE);

	my $old_focused = $self->{focused};
	$term->grab_focus;

	my ($but, $time);
	if ($button) {
		$but = $button->button;
		$time = $button->time;
	} else {
		$but = 3;
		$time = Gtk2::get_current_event_time(undef);
	}

	my $handler;
	$handler = $self->{context_menu}->signal_connect(deactivate => sub {
		$term->set("showing_context_menu", FALSE);
		# restore focus, unless term became the new director
		# or the focus changed to another
		$old_focused->grab_focus
			unless $term->get("role") == ROLE_DIRECTOR
				|| $self->{focused} != $term;
		# remove handler, we'll put one back when we need it
		$self->{context_menu}->signal_handler_disconnect($handler);
		return FALSE;
	});

	$self->{context_menu}->popup(undef, undef, undef, undef, $but, $time);
}

sub on_show_menu {
	my ($self, $show) = @_;

	$show ||= 0;
	return unless defined $self->{show_menus}
		&& $show != $self->{show_menus};

	my @size;
	if ($self->{term}[0]) {
		@size = (
			$self->{term}[0]->get_column_count,
			$self->{term}[0]->get_row_count
		);
	}

	if ($show) {
		$self->{menubar}->show;
	} else {
		$self->{menubar}->hide;
	}
	$self->{show_menus} = $show;
	$self->set_my_geometry;
	$self->resize_from_cols_rows(@size);
}

sub on_close {
	my ($self, $why, $rc, $term) = @_;

	$self->remove_highlights();
	if ($why eq 'button') {
		$self->close_host($term);
	} elsif ($rc && $term->age < 120) {
		# if there was an error and the term was opened
		# less then 2 minutes, show the return code
		Glib::Timeout->add(200, sub {
			$term->feed("\r\nCommand returned $rc "
				."(use close button to close terminal).\r\n");
		});
	} else {
		# delay, so can see some output, but then close
		Glib::Timeout->add(500, \&on_close_delayed, [ @_ ])
	}
	$self->_update_menus();
}

sub on_close_delayed {
	&close_host(@{$_[0]});
	return FALSE;
}

sub on_commit {
	my ($self, $char, $len, $obj) = @_;

	# filter some weird stuff out
	$char =~ s/\x1b\].*?\x1b\\//g;
	return unless ($char ne "");

	return unless ($obj->{role} == ROLE_DIRECTOR);
	# send to other terminals, if the origin is director, and
	# the destination is active
	# instead of block/unblock, we should block on all that are
	# not director
	for (@{$self->{term}}) {
		# active should be an attribute!!
		next if $obj == $_ || $_->{role} != 0;

		# block our handler from being run recursivly
		$_->feed_child($char);
	}
}

sub on_key_press {
	my ($self, $obj, $event) = @_;

	#return &ClusTerm::Utils::do_accelerators($self->{shortcuts},
	#	$event->keyval, $event->state, $obj);
}

sub on_role_change {
	my ($self, $obj) = @_;
	my $newrole = $obj->get("role");

	# state model for role changes:
	# single click:
	# 	ACTOR -> FREE
	# 	FREE -> DIRECTOR, if previously DIRECTOR
	# 	FREE -> ACTOR, otherwise
	# double click:
	# 	any -> DIRECTOR
	if ($newrole == ROLE_DIRECTOR
		&& (! $self->{director} || $self->{director} != $obj)) {
		$obj->signal_handler_unblock($obj->{userdata}{commit_handler});

		if ($self->{director}) {
			# make the old director a actor
			$self->{director}->set_role(ROLE_ACTOR);
			$self->{director}->signal_handler_block(
				$self->{director}->{userdata}{commit_handler});
		}

		$self->{director} = $obj;
	} elsif ($newrole == ROLE_FREE) {
		if ($self->{director} && $obj == $self->{director}) {
			$obj->signal_handler_block(
				$obj->{userdata}{commit_handler});
			$self->{prev_director} = $self->{director};
			$self->{director} = undef;
		}
	} elsif ($newrole == ROLE_ACTOR && $self->{prev_director}
		&& ! $self->{director} && $obj == $self->{prev_director}
		&& $obj->get("has_single_role_button")) {
		$obj->set_role(ROLE_DIRECTOR);
		$obj->signal_handler_unblock($obj->{userdata}{commit_handler});
		$self->{director} = $obj;
	}
	
}

sub on_focus_term {
	my ($self, $event, $obj) = @_;

	$self->{focused} = $obj;
	return FALSE;
}

sub set_role {
	my ($self, $obj, $role, $all) = &get_term_args(@_);
	return unless @{$self->{term}};
	$obj->set_role($role);

	# use the role change handler, but prevent from making
	# the old director director again
	$self->{prev_director} = undef;

	if ($all) {
		for (@{$self->{term}}) {
			next if $_ == $obj;
			$_->set_role(ROLE_ACTOR);
		}
	}

	$self->on_role_change($obj);
}

sub set_focus {
	my ($self, $obj, $jump) = &get_term_args(@_);
	return unless @{$self->{term}};

	unless ($jump) {
		$obj->grab_focus;
		if ($self->{box}->isa('Gtk2::Notebook')) {
			for(my $i = 0; $i < @{$self->{term}}; $i++) {
				next unless $self->{term}[$i] == $obj;
				$self->{box}->set_current_page($i);
			}
		}
		return;
	}

	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		if ($self->{term}[$i] == $obj) {
			$i = ($i + $jump) % @{$self->{term}};
			$self->{term}[$i]->grab_focus;
			# previous doesn't trigger an event? strange...
			$self->{focused} = $self->{term}[$i];
			$self->{box}->set_current_page($i)
				if $self->{box}->isa('Gtk2::Notebook');
			return;
		}
	}
}

sub copy {
	my ($self, $obj) = &get_term_args(@_);
	return unless @{$self->{term}};
	$obj->{term}->copy_clipboard;
}

sub paste {
	my ($self, $obj) = &get_term_args(@_);
	return unless @{$self->{term}};

	# this will go via the commit handler, so it is automatically send
	# to all actors if this object is a director
	$obj->{term}->paste_clipboard;
}

sub insert_hostname {
	my ($self, $obj) = &get_term_args(@_);
	return unless @{$self->{term}};

	if ($obj->get("role") == ROLE_DIRECTOR) {
		$obj->signal_handler_block(
			$obj->{userdata}{commit_handler});
		for (@{$self->{term}}) {
			next if $_->get("role") == ROLE_FREE;
			$_->feed_child($_->{name});
		}
		$obj->signal_handler_unblock(
			$obj->{userdata}{commit_handler});
	} else {
		$obj->feed_child($obj->{name});
	}
}

sub close_host {
	my ($self, $obj) = &get_term_args(@_);
	$_[0]->remove_highlights();

	$self->{director} = undef
		if ($self->{director} && $self->{director} == $obj);
	$self->{prev_director} = undef
		if ($self->{prev_director} && $self->{prev_director} == $obj);

	# find that object and remove it
	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		next unless ($self->{term}[$i] == $obj);

		splice @{$self->{term}}, $i, 1;
		$self->remove($obj);
		last;
	}

	unless (@{$self->{term}}) {
		Gtk2->main_quit
			if $self->{prefs}->get("exit-after-last");
		return;
	}

	if ($self->{director}) {
		$self->{director}->grab_focus;
	} elsif ($self->{term}[0]) {
		$self->{term}[0]->grab_focus;
	}
	if (@{$self->{term}}) {
		$self->set_resizable(TRUE) if (@{$self->{term}});
	}
	$self->_update_menus();
	return TRUE;
}

sub set_filename {
	$_[0]{last_filename} = $_[1];
}

sub save_session_launcher {
	my $self = shift;

	unless ($self->{session_template}) {
		# find template file
		my $file;
		for (@INC) {
			last if open $file, '<',
				catfile($_,
				"clusterm-session-template.desktop");
		}

		unless ($file) {
			$self->error("<span weight=\"bold\" size=\"larger\">"
				."Could not find the launcher template file."
				."</span>\n"
				."Please reinstall the clusterm application.");
			return;
		}

		local $/ = undef;
		$self->{session_template} = <$file>;
		close $file;
	}

	my @name = map { $_->get_name } @{$self->{term}};
	my $name = join(', ', @name);
	my $session = ! @name
		? 'empty'
		: join('-', @name);
	$session =~ s{[^-.\@\w;:,]}{}g;

	my $chooser = Gtk2::FileChooserDialog->new(
		'Save Session As Launcher...',
		$self, 'save',
		'gtk-cancel' => 'cancel',
		'gtk-save' => 'ok');
	if ($self->{last_filename}) {
		my ($name, $path) = fileparse($self->{last_filename});
		$chooser->set_current_folder($path);
		$chooser->set_current_name($name);
	} else {
		$chooser->set_current_name("clusterm-$session.desktop");
	}
retry:
	my $res = $chooser->run;
	my $filename = $chooser->get_filename;
	goto done if $res ne 'ok';

	my $arg = "-".$self->{type}." "
		.join(" ", map {
			my $a = $_->get_arg;
			$a =~ s/"/\\"/g;
			"\"".$a."\""
		} @{$self->{term}});

	my $s = $self->{session_template};
	$s =~ s{(?<!\%)\%n}{$name}g;
	$s =~ s{(?<!\%)\%a}{$arg}g;

	if (-f $filename) {
		unless ($self->{last_filename}
			&& $filename eq $self->{last_filename}) {
			unless ($self->question_yn(
				"<span weight=\"bold\" size=\"larger\">"
				."A file named \"%s\" already "
				."exists. Do you want to replace it?"
				."</span>\n"
				."The file already exists in \"%s\". "
				."Replacing it will overwite its contents.",
				fileparse($filename))) {
				goto retry;
			}
		}
	}

	my $file;
	unless (open($file, '>', $filename)
		&& print $file $s) {
		$self->error("<span weight=\"bold\" size=\"larger\">"
			."Could not write to the file."
			."</span>\n"
			."$filename: $!");
		close $file if $file;
		goto done;
	}
	close $file;

	$self->{last_filename} = $filename;

done:
	$chooser->destroy;
}

sub error {
	my $self = shift @_;
	my $text = sprintf(shift @_, @_);
	my $dialog = Gtk2::MessageDialog->new_with_markup($self,
		'destroy-with-parent',
		'error',
		'ok',
		$text);
	$dialog->run;
	$dialog->destroy;
}

sub question_yn {
	my $self = shift @_;
	my $text = sprintf(shift @_, @_);
	my $dialog = Gtk2::MessageDialog->new_with_markup($self,
		'destroy-with-parent',
		'question',
		'yes-no',
		$text);
	my $r = $dialog->run;
	$dialog->destroy;
	return $r eq 'yes';
}

sub preferences {
	my $self = shift;
	$self->{prefs}->show_all;
	$self->{prefs}->present;
}

my $about_dialog;
sub about {
	if ($about_dialog) {
		$about_dialog->present;
		return;
	}
	$about_dialog = Gnome2::About->new($title, $VERSION, $copyright,
		$about, [ $author ]);
	# if I set the icon in new, then I have problems with translators
	# showing up
	$about_dialog->set("logo", $icon);
	$about_dialog->set_default_icon($icon);
	$about_dialog->signal_connect(
		destroy => sub { $about_dialog = undef });
	$about_dialog->show_all;
}

sub get_term_args {
	my $self = shift @_;
	my $term;
	if (&isa_term($_[-1])) {
		$term = pop @_;
	} elsif (&isa_term($_[0])) {
		$term = shift @_;
	} else {
		$term = $self->{focused};
	}

	return ($self, $term, @_);
}

sub isa_term {
	return $_[0] && ref $_[0] && $_[0]->isa('ClusTerm::Term');
}

sub _create_box {
	my ($self, $size) = @_;

	if ($self->{type} =~ /^row/i) {
		$self->{type} = 'row';
		$self->{box} = Gtk2::HBox->new(0, 0);
		$self->{box}->set_spacing($self->{spacing}[0]);
		($self->{cols}, $self->{rows}) = ($size, 1);
	} elsif ($self->{type} =~ /^col/i) {
		$self->{type} = 'column';
		$self->{box} = Gtk2::VBox->new(0, 0);
		$self->{box}->set_spacing($self->{spacing}[1]);
		($self->{cols}, $self->{rows}) = (1, $size);
	} elsif ($self->{type} =~ /^pane/i) {
		$self->{type} = 'pane';
		return undef if ($size != 2);
		$self->{box} = Gtk2::HPaned->new();
		($self->{cols}, $self->{rows}) = ($size, 1);
	} elsif ($self->{type} =~ /^note/i) {
		$self->{type} = 'notebook';
		$self->{box} = Gtk2::Notebook->new();
		$self->{box}->set_tab_pos('bottom');
		$self->{box}->popup_enable();
		$self->{box}->set_show_border(FALSE);
		$self->{box}->set_scrollable(TRUE);
		($self->{cols}, $self->{rows}) = (1, 1);

		# calculate tab height by adding something to the notebook
		# sizing the notebook, and then removing it again
		my $w = Gtk2::Label->new();
		$w->set_size_request(120, 40);
		$self->{box}->append_page($w, 'Ng');
		$self->{box}->show_all();
		my $w_req = $w->size_request();
		my $note_req = $self->{box}->size_request();
		$self->{notebook_tab_height}
			= $note_req->height - $w_req->height - 4;
		$self->{notebook_tab_width}
			= $note_req->width - $w_req->width - 2;
		# 4 and 2 make the perfect layout, why?
		$self->{box}->remove_page(0);
	} elsif ($self->{type} =~ /^tab/i) {
		$self->{type} = 'table';
		# try to be the most square, and more horizental if needed
		# ceil x := 100 - int(100 - x)
		$self->{cols} = 100 - int(100 - sqrt($size));
		$self->{cols} = 1 unless ($self->{cols} > 1);
		$self->{rows} = 100 - int(100 - $size / $self->{cols});
	
		$self->{box} = Gtk2::Table->new($self->{rows}, $self->{cols},
			TRUE);
		map { $self->{box}->set_col_spacing($_-1, $self->{spacing}[0]) }
			(1..$self->{cols});
		map { $self->{box}->set_row_spacing($_-1, $self->{spacing}[1]) }
			(1..$self->{rows});
	}

	return $self->{box} ? TRUE : FALSE;
}

sub add_host {
	my ($self, $cmd, $role) = @_;
	$role = defined $role ? $role : ROLE_ACTOR;

	my $arg = $cmd;
	my $host = $cmd;
	if ($cmd =~ m{^/}) {
		$host =~ s{^.*/}{};
	} elsif ($cmd =~ /\s/) {
		$host =~ s/^\S*\s+//;
	} else {
		$cmd = $self->{prefs}->get('terminal-command');
		$cmd =~ s/\%h/$host/g;
	}
	$host =~ s{[^-.\@\s/\w;:,'"]}{}g;

	my $ud = { };
	local $_ = ClusTerm::Term->new($host, $role, $ud);
	push @{$self->{term}}, $_;
	$self->add($_, $#{$self->{term}});

	$_->set_arg($arg);
	$_->fork_command('/bin/sh', ['sh', '-c', 'exec '.$cmd], undef,
		'/', FALSE, FALSE, FALSE);
	$_->signal_connect_swapped(close => \&on_close, $self);
	$_->signal_connect_swapped(role_change => \&on_role_change, $self);
	$_->signal_connect_swapped(context_menu => \&context_menu, $self);
	$_->signal_connect_swapped(focus_in_event => \&on_focus_term, $self);

	$_->set('profile', $self->{prefs}->get('profile'));
	$self->{prefs}->link_properties('title-font-name', $_);
	$self->{prefs}->link_properties('director-color', $_);
	$self->{prefs}->link_properties('focused-color', $_);
	$self->{prefs}->link_properties('free-color', $_);
	$self->{prefs}->link_properties('show-close',
		$_, 'has-close-button');
	$self->{prefs}->link_properties('show-single-role',
		$_, 'has-single-role-button');
	$self->{prefs}->link_properties('close-on-disconnect',
		$_, 'close-on-disconnect');

	$ud->{commit_handler} = $_->signal_connect_swapped(
		commit => \&on_commit, $self);
	$_->signal_handler_block($ud->{commit_handler})
		if ($role != ROLE_DIRECTOR);
	$self->_update_menus();
	return $_;
}

sub add {
	my ($self, $obj, $idx) = @_;

	if ($self->{type} =~ /^(row|col)/i) {
		$self->{box}->pack_start($obj, TRUE, TRUE, 0);
	} elsif ($self->{type} =~ /^pane/i) {
		if ($idx == 0) {
			$self->{box}->add1($obj);
		} else {
			$self->{box}->add2($obj);
		}
	} elsif ($self->{type} =~ /^note/i) {
		$self->{box}->append_page($obj, $obj->{name});

		# notebook will consume a lot of CPU until the new page
		# has been shown, so quickly show it and then move back
		my $focus = 0;
		if ($self->{focused}) {
			for(my $i = 0; $i < @{$self->{term}}; $i++) {
				next unless $self->{term}[$i]
					== $self->{focused};
				$focus = $i;
				last;
			}
		}

		my $set;
		Glib::Timeout->add(10, sub {
			if ($set) {
				$self->{box}->set_current_page($focus);
				return 0;
			} else {
				$self->{box}->set_current_page($idx);
				$set = 1;
				return 1;
			}
		});
	} elsif ($self->{type} =~ /^tab/i) {
		$self->resize_box(1);
		$self->{cols} ||= 1;
		$self->{rows} ||= 1;
		my $r = int($idx / $self->{cols});
		my $c = $idx % $self->{cols};
		$self->{box}->attach_defaults($obj, $c, $c+1, $r, $r+1);
	}

	# focus if it is the only one
	$obj->grab_focus if ($idx == 0);
}

sub resize_box {
	my ($self, $add) = @_;
	my ($c, $r) = ($self->{cols}, $self->{rows});
	my $n = scalar @{$self->{term}};

	if (($add && $n > $c * $r)
		|| (!$add && ($n <= $c * ($r - 1) || $n <= ($c - 1) * $r))) {
		my @c = $self->{box}->get_children;
		unless (@c) {
			$self->{cols} = 0;
			$self->{rows} = 0;
			return;
		}

		map { $self->{box}->remove($_) } @c;

		$self->{cols} = 100 - int(100 - sqrt($n));
		$self->{rows} = 100 - int(100 - $n / $self->{cols});
		$self->{box}->resize($self->{rows}, $self->{cols});

		# put back in the order that they apear in $self->{term}
		my %c;
		@c{@c} = ( );
		@c = ( );
		map { push @c, $_ if exists $c{$_} } @{$self->{term}};
		map { $self->add($c[$_], $_) if exists $c{$c[$_]}; } (0..$#c);
	}
}

sub remove {
	my ($self, $obj) = @_;

	# remove hints
	my $geom = Gtk2::Gdk::Geometry->new;
	$self->set_geometry_hints($self, $geom, [ ]);
	my @size;
	@size = (
		$self->{term}[0]->get_column_count,
		$self->{term}[0]->get_row_count
	) if $self->{term}[0];

	if ($self->{type} =~ /^(row|col|pane|note)/i) {
		$self->{box}->remove($obj);
	} elsif ($self->{type} =~ /^tab/i) {
		$self->{box}->remove($obj);
		$self->resize_box;
	}

	$self->resize_from_cols_rows(@size);
	$self->set_my_geometry();
}

sub zoom_in {
	# zoom should be handled in Term
	my ($self) = @_;

	$self->{zoom} += .2;
	$self->set_term_zoom;
}

sub zoom_out {
	my ($self) = @_;

	$self->{zoom} -= .2 if $self->{zoom} >= .2;
	$self->set_term_zoom;
}

sub zoom_normal {
	my ($self) = @_;

	$self->{zoom} = 1;
	$self->set_term_zoom;
}

sub set_term_zoom {
	my ($self, $font) = @_;

	return unless @{$self->{term}};

	my @size = (
		$self->{term}[0]->get_column_count,
		$self->{term}[0]->get_row_count
	);

	for (@{$self->{term}}) {
		$_->set_zoom($self->{zoom});
	}
	$self->set_my_geometry;
	$self->resize_from_cols_rows(@size);
}

sub change_type {
	my ($self, $newtype, $radio) = @_;
	return if $newtype eq $self->{type} || ! $radio->get_active;

	# remove hints
	my $geom = Gtk2::Gdk::Geometry->new;
	$self->set_geometry_hints($self, $geom, [ ]);

	my @size = (
		$self->{term}[0]->get_column_count,
		$self->{term}[0]->get_row_count
	) if @{$self->{term}};

	my $focus = $self->{focused};
	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		$self->{box}->remove($self->{term}[$i]);
	}

	$self->{type} = $newtype;
	$self->_create_box(scalar @{$self->{term}});
	$self->set_contents($self->{box});
	return unless @{$self->{term}};

	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		$self->add($self->{term}[$i], $i);
	}

	$self->set_my_geometry;
	$self->resize_from_cols_rows(@size);
	# how to make it focus/set_current_page correctly??
	Glib::Timeout->add(50, sub { $self->set_focus($focus); return 0; })
}

sub highlight_diffs {
	my ($self) = @_;
	return unless @{$self->{term}} > 1;

	# remember diffs
	my $diffs = $self->{highlights};
	# remove old diffs
	$self->remove_highlights;

	my $gconf = Gnome2::GConf::Client->get_default;
	my $profile = $self->{prefs}->get('profile') || 'Default';
	my $word_chars = $gconf->get_string(
		"/apps/gnome-terminal/profiles/$profile/word_chars", FALSE);
	my $split_re = $word_chars ? qr/(?:[^$word_chars\s]+|\s+)/ : qr/\s+/;
	
	my @wordspaces;
	my @words;
	my @attrs;
	my @toprow;
	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		my ($text, $attrs) = $self->{term}[$i]->get_active_text();

		# restore attributes from before highlighting
		# (the changes from remove_highlights have not yet
		# percolated to the values we just retrieved)
		if ($diffs && $diffs->[$i]) {
			for my $d (@{$diffs->[$i]}) {
				next unless $d->{found};
				# note that remove_highlights has updated
				# the offset
				my $o = $d->{off};
				my $e = $o + $#{$d->{attrs}};
				@{$attrs}[$o..$e] = @{$d->{attrs}};
			}
		}

		push @attrs, $attrs;
		push @toprow, $attrs->[0]{row};
		my $cols = $self->{term}[$i]->get_column_count();

		# reverse to prefer matches at the end
		my @a = reverse @{ split_rc($split_re, $text, $cols) };
		map { $a[$_]{idx} = $_ } (0..$#a);
		push @wordspaces, \@a;
		# remove spaces
		push @words, [ grep { $_->{t} =~ /\S/ } @{$wordspaces[-1]} ];
	}

	my $win = Gtk2::Gdk::Window->new(
		$self->window(), {
		wclass => 'only',
		type => 'child',
		x => 0,
		y => 0,
		width => 32767,
		height => 32767,
	});
	$win->show;
	my $cursor = Gtk2::Gdk::Cursor->new('watch');

#print "Lengths: ", join(", ", map { scalar @$_ } @words), "\n";
my $start = Time::HiRes::time;
	my @lcss;
	eval {
		my $count = 0;
		# timeout search, to prevent annoying delays
		local $SIG{ALRM} = sub {
			if ($count > 20) {
				die "timeout\n";
			} elsif (! $count) {
				$win->set_cursor($cursor);
			}
			Glib::MainContext->default()->iteration(TRUE);
			$count++;
			alarm(.5);
		};
		alarm(.5);
		@lcss = lcss(@words, sub { $_->{t} });
		alarm(0);
	};
	if ($@) {
		print STDERR "Did not find differences: $@";
		goto nodiff;
	}
my $end = Time::HiRes::time;
printf "LCSS: %d, in %.2f\n", scalar @lcss, $end - $start;

	# get all the differences, and the matching attributes
	# add an explicit match just past the end of each array
	push @lcss, [ map { scalar @{$words[$_]} } (0..$#words) ];
	my @diffs;
	for my $i (0..$#words) {
		my $match;
		my $first_match = -1;
		for(my $l = 0; $l < @lcss; $l++, $first_match = $match) {
			$match = $lcss[$l][$i];
			next if $match == $first_match + 1;

			# remember it was reversed
			my $first_diff = $words[$i][$first_match + 1];
			my $last_diff = $words[$i][$match - 1];

			my $first_idx = $first_diff->{idx};
			my $last_idx = $last_diff->{idx};
			my $diff_text = join("", map { $_->{t} } reverse
				@{$wordspaces[$i]}[$first_idx..$last_idx]);

			$first_idx = $last_diff->{off};
			$last_idx = $first_diff->{off}
				+ length($first_diff->{t}) - 1;
			my @diff_attr = @{$attrs[$i]}[$first_idx..$last_idx];
			push @{$diffs[$i]}, {
				text => $diff_text,
				attrs => \@diff_attr,
				col => $last_diff->{c},
				row => $last_diff->{r},
				toprow => $attrs[$i][0]{row},
				off => $last_diff->{off},
			};
		}
	}
	$self->{highlights} = \@diffs;

	my $fg = $self->{prefs}->get('highlight-foreground');
	my $bg = $self->{prefs}->get('highlight-background');
	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		next unless $diffs[$i];
		$self->_show_text_attr($self->{term}[$i], $diffs[$i],
			{ fg => $fg, bg => $bg });
	}

nodiff:
	$win->hide() if $win;
}

sub remove_highlights {
	my ($self) = @_;
	return unless $self->{highlights};

	for(my $i = 0; $i < @{$self->{term}}; $i++) {
		next unless $self->{highlights}[$i];

		# have to adjust for possible scrolling
		my ($text, $attrs) = $self->{term}[$i]->get_active_text();
		my $toprow = $attrs->[0]{row}
			- $self->{highlights}[$i][0]{toprow};

		my @high;
		my $pos = 0;
		for my $high (reverse @{$self->{highlights}[$i]}) {
			my %d = %$high;
			my $found = $pos;
			while(($found = index $text, $d{text}, $found)
				>= 0) {
				my $olda = $d{attrs}[0];
				my $newa = $attrs->[$found];
				# text should be at the same position
				# in the scrollback buffer (that
				# is the position reported in column, row)
				unless ($olda->{column} == $newa->{column}
					&& $olda->{row} == $newa->{row}) {
					$found++;
					next;
				}

				# adjust for scrolling
				$d{row} -= $toprow;
				$d{toprow} += $toprow;

				# update for use by highlight_diffs
				$high->{off} = $found;
				$high->{found} = 1;

				push @high, \%d;
				$pos = $found + length($d{text});
				last;
			}
		}

		$self->_show_text_attr($self->{term}[$i], \@high);
	}
	$self->{highlights} = undef;
}

sub _show_text_attr {
	my ($self, $term, $texts, $override) = @_;

	return unless @$texts;

	my $columns = $term->get_column_count();
	my ($col, $row) = $term->get_cursor_position();

	# make a local copy of the array, so we don't change the original
	# array
	$texts = [ @$texts ];
	for(my $j = 0; $j < @$texts; $j++) {
		my $a = $texts->[$j];
		my $text = $a->{text};

		if (0 && $text =~ /\n/) {
			# split into blobs on single lines
			my $col = $a->{col};
			my $row = $a->{row};
			my $pos = 0;
			my @new;
			for my $t (split /(\n)/, $text) {
				if ($t =~ /\n/) {
					$pos += length($t);
					next;
				}

				my $end = $pos + length($t) -1;
				push @new, {
					col => $col,
					row => $row,
					text => $t,
					toprow => $a->{toprow},
					attrs => [ @{$a->{attrs}}[$pos..$end] ],
				};
				$col = 0;
				$row++;
				$pos += length($t);
			}
			splice @$texts, $j, 1, @new;
			redo;
		}

		# don't show text past the end of the row, or cut
		# to the end of the row
		#print "dont show \"$text\"\n" if $a->{col} > $columns;
		#next if $a->{col} > $columns;
		#$text = substr($text, 0, $columns - $a->{col} + 1)
		#	if $a->{col} + length($text) - 1 > $columns;

		$term->output_set_cursor($a->{col}, $a->{row});
		$term->output($a->{text}, $a->{attrs}, $override);
	}

	$term->output_set_cursor($col, $row - $texts->[0]{toprow});
	$term->output_flush();
}


1;
