##
## Copyright (c) 2005-2007 by Rafael D'Halleweyn
##
## Licensed under Version 2 of the GPL
##

# the referenced signals and properties have to be added when
# the type is registered. Properties get connected at that time
# also.
# Then when an object is created, the signals are connected.
#
# $object->notify($prop_name), emits notify signal for the prop on this obj

# oh, should also change the notify signal for the attributes!!

package ClusTerm::GlibSubclass;
our $VERSION = '0.2';
use strict;
use warnings;
use Carp;
use Glib qw(TRUE FALSE);
use Gnome2::GConf;
use ClusTerm::Utils qw(color_gdk color_string);

my %subclass;

*new = \&Glib::Object::new;

sub import {
	# if imported by classes using classes which use us, do nothing
	return unless @_ > 1;

	my ($self, $superclass, %arg) = @_;
	my $class = $arg{class} || caller;

	# make sure that all signals use - (not _)
	for (keys %{$arg{signals}}) {
		my $orig = $_;
		if (s/_/-/g) {
			$arg{signals}{$_} = $arg{signals}{$orig};
			delete $arg{signals}{$orig};
		}

	}

	# remember for later
	my ($src_name, $refed);
	while(($src_name, $refed) = each %{$arg{references}}) {
		my @refed = @$refed;
		my $type = shift @refed;
		my $prop_list = [ Glib::Object::list_properties($type) ];

		for my $src (@refed) {
			my $dst = $src;
			my ($trans_to, $trans_from);
			($src, $dst, $trans_to, $trans_from) = @$src
				if (ref $src);
			$dst ||= $src;

			if ($type->can($src)) {
				no strict 'refs';
				my $name = $class.'::'.$dst;
				my $key = $src_name;

				if ($class->can($dst)) {
					# this will not happen,
					# the methods in the class are
					# probably not seen yet
					# (we are in BEGIN of that package)
					carp("$class already has a method ",
						"named \'$dst\'");
					next;
				}
				*$name = sub {
					my $self = shift @_;
					$self->{$key}->$src(@_);
				};
				next;
			}

			$src =~ s/_/-/g;
			$dst =~ s/_/-/g;

			# if the signal is already defined, don't bother
			# looking it up
			if (exists $arg{signals}{$dst}) {
				# only remember to connect
				goto store_connect;
			} elsif (my $sig = Glib::Object::signal_query(
					$type, $src)) {
				for (qw(return_type param_types)) {
					$arg{signals}{$dst}{$_} = $sig->{$_};
				}
				$arg{signals}{$dst}{signal_flags}
					= $sig->{flags};
store_connect:
				$subclass{$class}{signal}{$dst} = {
					src_name => $src_name,
					src_sig_name => $src,
				};
			} elsif (my ($prop) = grep { $_->{name} eq $src }
					@$prop_list) {
				my $newprop = &_connect_properties($prop,
					$src_name, $src, $dst);
				push @{$arg{properties}}, $newprop
					if ($newprop);

				$subclass{$class}{property}{$dst} = {
					src_name => $src_name,
					src_prop_name => $src,
				};
			} else {
				carp("$type does not have any attribute, "
					."signal or method named '$src'");
			}
		}
	}

	for my $p (@{$arg{properties}}) {
		my $prop = $p->{pspec}->{name};

		if ($p->{pspec}{type} eq 'Gtk2::Gdk::Color') {
			# provide a default set for Gtk2::Gdk::Color properties
			$p->{set} ||= sub {
				$_[0]->{$prop} = &color_gdk($_[1]);
			};
		}

		if (exists $arg{gconf}{$prop}) {
			$subclass{$class}{gconf}{$prop} = $p;
			$p->{orig_get} = $p->{get};
			$p->{orig_set} = $p->{set};
			$p->{default} = $arg{gconf}{$prop};
			$p->{get} = sub {
				$_[0]->_gconf_get($class, $p, @_[1..$#_])
			};
			$p->{set} = sub {
				$_[0]->_gconf_set($class, $p, @_[1..$#_])
			};
			# should overwrite set and get handlers
		}
	}

	# do what Glib::Object::Subclass does (no way to reuse it,
	# since the caller would be wrong, we would define the wrong class)
	Glib::Type->register_object($superclass, $class, %arg);

	# we want to be a superclass of the new class, so we can put in
	# a default new
	no strict 'refs';
	unshift @{ $class."::ISA" }, __PACKAGE__;
}

sub subclass_connect {
	my $object = shift;
	my %arg = @_;
	my $class = caller;
	return $object unless exists $subclass{$class};

	$object->_connect_gconf($class, $arg{gconf_key})
		if ($arg{gconf_key});

	my ($sig, $prop, $info);
	while(($sig, $info) = each %{$subclass{$class}{signal}}) {
		my $n = $info->{src_name};
		croak("$object does not contain '$n'")
			unless ($object->{$n});
		$object->{$n}->signal_connect(
			$info->{src_sig_name} => _signal_closure($object, $sig,
				$info->{trans}));
	}
	while(($prop, $info) = each %{$subclass{$class}{property}}) {
		my $n = $info->{src_name};
		croak("$object does not contain '$n'")
			unless ($object->{$n});
		$object->link_properties($prop,
			$object->{$n}, $info->{src_prop_name}, TRUE);
	}

	return $object;
}

my %notified;
sub link_properties {
	my ($obj1, $attr1, $obj2, $attr2, $bidir) = @_;

	my ($prop1, $prop2);
	$attr2 ||= $attr1;
	croak("Object $obj1 does not have an attribute '$attr1'")
		unless (($prop1) = grep { $_->{name} eq $attr1 }
			$obj1->list_properties);
	croak("Object $obj2 does not have an attribute '$attr2'")
		unless (($prop2) = grep { $_->{name} eq $attr2 }
			$obj2->list_properties);

	&_add_notify($obj1, $attr1, $obj2, $attr2, $prop2);
	&_add_notify($obj2, $attr2, $obj1, $attr1, $prop1) if ($bidir);

	# now copy the current value by faking a notify event
	&_notify_handler($obj1, { name => $attr1 });
}

sub _add_notify {
	my ($obj, $attr, $dst_obj, $dst_attr, $dst_prop) = @_;
	my $handler;

	unless ($notified{$obj} && $notified{$obj}{$attr}) {
		$obj->notify($attr);
	}

	if ($notified{$obj}{_handler}) {
		$handler = $notified{$obj}{_handler};
	} else {
		$handler = $obj->signal_connect(notify => \&_notify_handler);
		$notified{$obj}{_handler} = $handler;
	}

	push @{$notified{$obj}{$attr}}, {
		dst_obj => $dst_obj,
		dst_attr => $dst_attr,
		dst_prop => $dst_prop,
		handler => $handler,
	};
	return $handler;
}

sub _notify_handler {
	my ($obj, $pspec) = @_;
	my $attr = $pspec->{name};

	return unless exists $notified{$obj} && exists $notified{$obj}{$attr};

	# get value
	my $val = $obj->get($attr);
	return unless defined $val;
	my $src_type = ref $val ? ref $val : "";

	# send to listeners
	my %done;
	my @listeners = @{$notified{$obj}{$attr}};
	$done{$obj}{$attr} = TRUE;

	# instead of letting the notify signals cascading all over the
	# place, we will do the work ourselves, that means we disable
	# any other notify handlers that we find on our path, and also
	# that we notify our listener's listeners.
	while(my $not = shift @listeners) {
		my $dst = $not->{dst_obj};
		$attr = $not->{dst_attr};
		next if $done{$dst}{$attr}++;
		my $dst_type = $not->{dst_prop}{type};
		$dst_type = '' if $dst_type =~ /^Glib::(String|Boolean)$/;

		my $blocked_handler;
		if (exists $notified{$dst} && exists $notified{$dst}{$attr}) {
			$blocked_handler = $notified{$dst}{_handler};
			$dst->signal_handler_block($blocked_handler);
			push @listeners, @{$notified{$dst}{$attr}};
		}

		if ($src_type eq $dst_type) {
			$dst->set($attr, $val);
		} else {
			croak("Don't know how to convert from '$src_type' "
				."to '$dst_type'");
		}

		$dst->signal_handler_unblock($blocked_handler)
			if ($blocked_handler);
	}
}

my %param_map = (
	# $_[0] source type
	# $_[1] name
	# $_[2] nick
	# $_[3] blurb
	# $_[4] flags
	Glib::Param::Int::
		=> sub { Glib::ParamSpec->IV($_[1], $_[2], $_[3],
			-~0, ~0, $_[0]->get_default_value, $_[4]) },
	Glib::Param::String::
		=> sub { Glib::ParamSpec->string($_[1], $_[2], $_[3],
			$_[0]->get_default_value, $_[4]) },
	Glib::Param::Boolean::
		=> sub { Glib::ParamSpec->boolean($_[1], $_[2], $_[3],
			$_[0]->get_default_value, $_[4]) },
	Glib::Param::Boxed::
		=> sub { Glib::ParamSpec->boxed($_[1], $_[2], $_[3],
			$_[0]->{type}, $_[4]) },
);
sub _connect_properties {
	my ($prop, $src_name, $src, $dst, $trans_to, $trans_from) = @_;
	my %newprop;
	# lame nick
	(my $nick = $dst) =~ s/[-_]/ /g;

	# oho, need to remove G_PARAM_STATIC_NAME from flags!!
	my $flags = $prop->{flags} & [qw(readable writable
		construct construct_only lax_validation)];

	for my $t (keys %param_map) {
		if ($prop->isa($t)) {
			$newprop{pspec} = $param_map{$t}->($prop,
				$dst, $nick, $prop->{descr}, $flags);
			last;
		}
	}

	if ($newprop{pspec}) {
	} elsif ($prop->isa('Glib::Object')) {
		$newprop{pspec} = Glib::ParamSpec->object($dst, $nick,
			$prop->{descr}, ref $prop, $flags);
	} else {
		croak("Don't know how to copy the property '$prop->{type}'");
	}

	return \%newprop;
}

sub _signal_closure {
	my ($obj, $sig, $trans) = @_;

	if ($trans) {
		return sub {
			shift;
			$obj->signal_emit($sig, $obj->trans(@_));
		};
	} else {
		return sub {
			shift;
			$obj->signal_emit($sig, @_);
		};
	}
}

sub methods {
	my ($class, $types) = @_;
	$class = ref $class || $class;
	$types ||= '';
	my %classes_seen;
	my %methods;
	my @class = ($class);

	no strict 'refs';
	while ($class = shift @class) {
		next if $classes_seen{$class}++;
		unshift @class, @{"${class}::ISA"} if $types eq 'all';
		# Based on methods_via() in perl5db.pl
		for my $method (grep {not /^[(_]/ and 
					defined &{${"${class}::"}{$_}}} 
				keys %{"${class}::"}) {
			$methods{$method} = wantarray
				? undef : $class->can($method); 
		}
	}
      
	wantarray ? keys %methods : \%methods;
}

sub _connect_gconf {
	my ($object, $class, $key) = @_;

	# this will fail if my sub-class also wants to do gconf
	$key =~ s{/+$}{};
	my $gconf = Gnome2::GConf::Client->get_default;
	$gconf->add_dir($key, 'preload-onelevel');
	$key .= '/';

	while(my ($prop, $pinfo) = each %{$subclass{$class}{gconf}}) {
		# make sure default values are set
		my $val = $gconf->get($key.$prop);
		if (! defined $val && defined $pinfo->{default}) {
			$val = $pinfo->{default};
			_gconf_set_convert($gconf, $key.$prop,
				$pinfo->{pspec}, $val);
		} else {
			$val = _gconf_get_convert($gconf, $key.$prop,
				$pinfo->{pspec});
			if ($pinfo->{orig_set}) {
				&{$pinfo->{orig_set}}($object, $val);
			} else {
				$object->{$prop} = $val;
			}
		}

		$object->{"_gconf_${class}_prop"}{$prop} = $pinfo;
	}

	$object->{"_gconf_${class}"} = $gconf;
	$object->{"_gconf_${class}_dir"} = $key;
	$object->{"_gconf_${class}"}->signal_connect_swapped(
		value_changed => \&_gconf_value_changed, [ $object, $class ]);
}

sub _gconf_get_convert {
	my ($gconf, $key, $pspec) = @_;

	my $val;
	if ($pspec->{type} eq 'Glib::Boolean') {
		$val = $gconf->get_bool($key);
	} elsif ($pspec->{type} eq 'blah') {
		$val = $gconf->get_int($key);
	} else {
		$val = $gconf->get_string($key);
		$val = &color_gdk($val)
			if $pspec->{type} eq 'Gtk2::Gdk::Color';
	}

	return $val;
}

sub _gconf_set_convert {
	my ($gconf, $key, $pspec, $val) = @_;
	return unless defined $val;

	if ($pspec->{type} eq 'Glib::Boolean') {
		return $gconf->set_bool($key, $val);
	} elsif ($pspec->{type} eq 'blah') {
		return $gconf->set_int($key, $val);
	}

	if ($pspec->{type} eq 'Gtk2::Gdk::Color') {
		$val = &color_string($val);
	}

	$gconf->set_string($key, $val);
}

sub _gconf_value_changed {
	my ($object, $class) = @{ shift @_ };
	my ($key, $gptr, $client) = @_;

	my $dir = $object->{"_gconf_${class}_dir"};
	return unless $key =~ s{^\Q$dir\E}{};
	my $prop = $key;

	unless (exists $object->{"_gconf_${class}_prop"}{$prop}) {
		carp("Do not know about a key \"$key\"");
		return;
	}
	my $pinfo = $object->{"_gconf_${class}_prop"}{$prop};

	if (exists $object->{"_gconf_${class}"}) {
		my $val = _gconf_get_convert($object->{"_gconf_${class}"},
			$object->{"_gconf_${class}_dir"}.$prop,
			$pinfo->{pspec});

		if ($pinfo->{orig_set}) {
			&{$pinfo->{orig_set}}($object, $val);
		} else {
			$object->{$prop} = $val;
		}
	}

	# we should only do a notify if gconf is the source of the change
	# not that we made the change ourselves...
	$object->signal_emit('notify', $pinfo->{pspec});
}

sub _gconf_get {
	my ($object, $class, $pinfo) = @_;
	my $name = $pinfo->{pspec}{name};

	if ($pinfo->{orig_get}) {
		return &{$pinfo->{orig_get}}($object);
	} else {
		return $object->{$name};
	}
}

sub _gconf_set {
	my ($object, $class, $pinfo, $value) = @_;
	my $name = $pinfo->{pspec}{name};

	if (exists $object->{"_gconf_${class}"}) {
		&_gconf_set_convert($object->{"_gconf_${class}"},
			$object->{"_gconf_${class}_dir"}.$name,
			$pinfo->{pspec}, $value);
		return;
	}
	
	if ($pinfo->{orig_set}) {
		return &{$pinfo->{orig_set}}($object, $value);
	} else {
		$object->{$name} = $value;
	}
}

1;
