##
## Copyright (c) 2005-2007 by Rafael D'Halleweyn
##
## Licensed under Version 2 of the GPL
##

package ClusTerm::Preferences;

use strict;
use warnings;
use Glib qw(TRUE FALSE);
use Gtk2;
use ClusTerm::Utils qw(color_gdk);

use ClusTerm::GlibSubclass Gtk2::Dialog::,
	properties => [
		{
			pspec => Glib::ParamSpec->string(
				'profile', 'Profile',
				'GNOME Terminal Profile',
				'Default',
				[qw/readable writable/]),
			set => \&set_profile,
		},
		{
			pspec => Glib::ParamSpec->string(
				'layout', 'Layout',
				'Default Layout',
				'Table',
				[qw/readable writable/]),
			set => \&set_layout,
		},
		{
			pspec => Glib::ParamSpec->string(
				'highlight-foreground', 'Highlight Foreground',
				'Highlight Foreground Color',
				'Black',
				[qw/readable writable/]),
			set => \&set_highlight_foreground,
		},
		{
			pspec => Glib::ParamSpec->string(
				'highlight-background', 'Highlight Background',
				'Highlight Background Color',
				'Blue',
				[qw/readable writable/]),
			set => \&set_highlight_background,
		},
	],
	references => {
		close_on_disconnect_button => [ Gtk2::CheckButton::,
			[ 'active' => 'close-on-disconnect' ],
		],
		exit_after_last_button => [ Gtk2::CheckButton::,
			[ 'active' => 'exit-after-last' ],
		],
		show_menubar_button => [ Gtk2::CheckButton::,
			[ 'active' => 'show-menubar' ],
		],
		title_font_button => [ Gtk2::FontButton::,
			[ 'font-name' => 'title-font-name' ],
		],
		director_color_button => [ Gtk2::ColorButton::,
			[ 'color' => 'director-color' ],
		],
		focused_color_button => [ Gtk2::ColorButton::,
			[ 'color' => 'focused-color' ],
		],
		free_color_button => [ Gtk2::ColorButton::,
			[ 'color' => 'free-color' ],
		],
		show_close_button => [ Gtk2::CheckButton::,
			[ 'active' => 'show-close' ],
		],
		show_single_role_button => [ Gtk2::CheckButton::,
			[ 'active' => 'show-single-role' ],
		],
		command_entry => [ Gtk2::Entry::,
			[ 'text' => 'terminal-command' ],
		],
	},
	gconf => {
		'profile' => 'Default',
		'layout' => 'Table',
		'show-menubar' => 1,
		'close-on-disconnect' => 1,
		'exit-after-last' => 1,
		'terminal-command' => 'ssh %h',
		'title-font-name' => 'Sans Bold 10',
		'director-color' => &color_gdk('#c90000'),
		'focused-color' => &color_gdk('#0076c9'),
		'free-color' => &color_gdk('#e6d6a1'),
		'show-close' => 1,
		'show-single-role' => 0,
		'highlight-foreground' => 'White',
		'highlight-background' => 'Blue',
	};

my @layout = qw(Table Row Column Notebook);
my @color = qw(Black Red Green Yellow Blue Magenta Cyan White);

sub new {
	my ($class, $app, $icon) = @_;
	my $self = Glib::Object::new(__PACKAGE__);

	$self->{gconf} = Gnome2::GConf::Client->get_default;
	$self->{gconf}->add_dir('/apps/gnome-terminal/global/profile_list',
		'preload-onelevel');

	$self->set_title('Preferences');
	$self->set_default_icon($icon) if ($icon);
	$self->set('has-separator', FALSE);
	$self->add_button('gtk-close', 1);
	$self->set('resizable', FALSE);

	$self->{tips} = Gtk2::Tooltips->new();

	my $notebook = Gtk2::Notebook->new();
	$notebook->set_tab_pos('top');
	$notebook->set_show_border(TRUE);
	$notebook->set_border_width(6);
	$self->vbox->add($notebook);

	$notebook->append_page(
		&_table_wrap('<b>General</b>', $self->_new_general),
		'General');
	$notebook->append_page(
		&_table_wrap('<b>Title Bar</b>', $self->_new_titlebar),
		'Title Bar');
	$notebook->append_page(
		&_table_wrap('<b>Highlighting</b>', $self->_new_highlighting),
		'Highlighting');

	# default values
	$self->set(
		show_menubar => 1,
	);
	$self->subclass_connect(gconf_key => "/apps/clusterm");
	$self->{gconf}->signal_connect_swapped(
		value_changed => \&_update_profiles, $self);
	return $self;
}

sub show_all {
	my ($self) = @_;

	$self->SUPER::show_all;
}

sub set_layout {
	my ($self, $layout) = @_;

	return unless defined $layout;
	$self->{layout} = $layout;
	return unless $self->{layout_combo};
	for(my $i = 0; $i < @layout; $i++) {
		next unless $layout eq $layout[$i];
		$self->{layout_combo}->set_active($i);
		return;
	}

	$self->{layout_combo}->set_active(0);
	$self->set('layout' => $layout[0]);
}

sub _changed_layout {
	my ($self) = @_;
	$self->{layout} = $self->{layout_combo}->get_active_text;
	$self->set('layout' => $self->{layout});
}

sub set_highlight_foreground {
	my ($self, $color) = @_;

	return unless defined $color;
	$self->{'highlight-foreground'} = $color;
	return unless $self->{highlight_foreground_combo};
	for(my $i = 0; $i < @color; $i++) {
		next unless $color eq $color[$i];
		$self->{highlight_foreground_combo}->set_active($i);
		return;
	}

	$self->{highlight_foreground_combo}->set_active(0);
	$self->set('highlight-foreground' => $color[7]);
}

sub _changed_highlight_foreground {
	my ($self) = @_;
	$self->{'highlight-foreground'}
		= $self->{highlight_foreground_combo}->get_active_text;
	$self->set('highlight-foreground' => $self->{'highlight-foreground'});
}

sub set_highlight_background {
	my ($self, $color) = @_;

	return unless defined $color;
	$self->{'highlight-background'} = $color;
	return unless $self->{highlight_background_combo};
	for(my $i = 0; $i < @color; $i++) {
		next unless $color eq $color[$i];
		$self->{highlight_background_combo}->set_active($i);
		return;
	}

	$self->{highlight_background_combo}->set_active(4);
	$self->set('highlight-background' => $color[4]);
}

sub _changed_highlight_background {
	my ($self) = @_;
	$self->{'highlight-background'}
		= $self->{highlight_background_combo}->get_active_text;
	$self->set('highlight-background' => $self->{'highlight-background'});
}

sub _update_profiles {
	my ($self) = @_;

	# prevent this from generating a cascade of changes
	$self->{profile_combo}->signal_handler_block($self->{profile_sigid});
	my $profile = $self->{profile};
	my $found = 0;
	map { $self->{profile_combo}->remove_text(0) } @{$self->{profiles}};

	# get the current gnome-terminal profiles
	$self->{profiles} = $self->{gconf}->get_list(
		'/apps/gnome-terminal/global/profile_list');

	$self->{profiles} = [ ] unless $self->{profiles};

	for(my $i = 0; $i < @{$self->{profiles}}; $i++) {
		$self->{profile_combo}->append_text($self->{profiles}[$i]);
		if ($profile && $profile eq $self->{profiles}[$i]) {
			$self->{profile_combo}->set_active($i);
			$found = 1;
		}
	}

	$self->{profile_combo}->signal_handler_unblock($self->{profile_sigid});

	if (! $found && $profile && @{$self->{profiles}}) {
		# our profile has been deleted, pick the first one that is left
		$self->set_profile($self->{profiles}[0]);
		$self->set('profile' => $self->{profile});
	}
}

sub set_profile {
	my ($self, $profile) = @_;

	unless ($profile) {
		$profile = $self->{profiles}[0];
		$self->{profile} = $profile;
		$self->{profile_combo}->set_active(0);
		return;
	}

	$self->{profile} = $profile;
	return unless $self->{profile_combo};
	for(my $i = 0; $i < @{$self->{profiles}}; $i++) {
		next unless $profile eq $self->{profiles}[$i];
		$self->{profile_combo}->set_active($i);
		return;
	}
}

sub _changed_profile {
	my ($self) = @_;
	my $new_profile = $self->{profile_combo}->get_active_text;
	return if $new_profile && $self->{profile}
		&& $new_profile eq $self->{profile};
	$self->{profile} = $new_profile;
	$self->set('profile' => $self->{profile});
}

sub _changed_command {
	my ($self, $event, $entry) = @_;

	my $text = $entry->get_text();

	$text =~ s/(?:\$host|\$\{host\}|\%s|\%h)/\%h/;
	$text .= ' %h' unless $text =~ /\%h/;
	$entry->set_text($text);

	$self->{terminal_command} = $text;
	$self->set('terminal-command' => $self->{terminal_command});

	return FALSE;
}

sub _new_label {
	my ($text) = @_;
	my $label = Gtk2::Label->new($text);
	$label->set_use_markup(TRUE);
	$label->set_use_underline(TRUE);
	$label->set_justify('left');
	$label->set_alignment(0.0, 0.5);
	return $label;
}

sub _table_wrap {
	my ($title, $w) = @_;

	my $table = Gtk2::Table->new(2,2);
	$table->set_border_width(12);
	$table->set_col_spacings(6);
	$table->set_row_spacings(6);
	$table->attach(&_new_label($title), 0, 2, 0, 1,
		['fill'], ['fill'], 0, 0);
	$table->attach(&_new_space(6), 0, 1, 1, 2,
		['fill'], ['fill'], 0, 0);
	$table->attach($w, 1, 2, 1, 2,
		['fill'], ['fill'], 0, 0);

	return $table;
}

sub _new_space {
	my ($w, $h) = @_;
	my $space = Gtk2::Label->new('');
	$space->set_size_request(defined $w ? $w : -1, defined $h ? $h : -1);
	return $space;
}

sub _new_general {
	my $self = shift;
	my $gen = Gtk2::Table->new(6, 2);
	$gen->set_col_spacings(6);
	$gen->set_row_spacings(6);

	my $lab = &_new_label('Default _layout:');
	$gen->attach_defaults($lab, 0, 1, 0, 1);
	$self->{layout_combo} = Gtk2::ComboBox->new_text;
	$lab->set_mnemonic_widget($self->{layout_combo});
	$self->{layout_combo}->signal_connect_swapped(
		changed => \&_changed_layout, $self);
	map { $self->{layout_combo}->append_text($_) } @layout;
	$gen->attach_defaults($self->{layout_combo}, 1, 2, 0, 1);

	$lab = &_new_label('GNOME Terminal _profile:');
	$gen->attach_defaults($lab, 0, 1, 1, 2);
	$self->{profile_combo} = Gtk2::ComboBox->new_text;
	$lab->set_mnemonic_widget($self->{profile_combo});
	$self->{profile_sigid} = $self->{profile_combo}->signal_connect_swapped(
		changed => \&_changed_profile, $self);
	$self->_update_profiles;
	$gen->attach_defaults($self->{profile_combo}, 1, 2, 1, 2);

	$self->{close_on_disconnect_button} = Gtk2::CheckButton->new(
		'Close terminal when disconnected');
	$gen->attach_defaults($self->{close_on_disconnect_button}, 0, 2, 2, 3);

	$self->{exit_after_last_button} = Gtk2::CheckButton->new(
		'_Exit after last terminal has closed');
	$gen->attach_defaults($self->{exit_after_last_button}, 0, 2, 3, 4);

	$self->{show_menubar_button} = Gtk2::CheckButton->new(
		'Show _menubar by default');
	$gen->attach_defaults($self->{show_menubar_button}, 0, 2, 4, 5);

	$lab = &_new_label('C_ommand:');
	$gen->attach_defaults($lab, 0, 1, 5, 6);
	$self->{command_entry} = Gtk2::Entry->new();
	$lab->set_mnemonic_widget($self->{command_entry});
	$self->{command_entry}->signal_connect_swapped(
		focus_out_event => \&_changed_command, $self);
	$gen->attach_defaults($self->{command_entry}, 1, 2, 5, 6);

	$self->{tips}->set_tip($self->{command_entry},
		"The command that will be\nexecuted for each new host.\n"
		."The text '%h' will be replaced\nwith the hostname.",
		"Some more text?");

	return &_align_left($gen);
}

sub _new_titlebar {
	my $self = shift;
	my $tit = Gtk2::Table->new(6, 2);
	$tit->set_col_spacings(6);
	$tit->set_row_spacings(6);

	my $lab = &_new_label('_Title font:');
	$tit->attach_defaults($lab, 0, 1, 0, 1);
	$self->{title_font_button} = Gtk2::FontButton->new;
	$lab->set_mnemonic_widget($self->{title_font_button});
	$self->{title_font_button}->set_title('Pick title font');
	$self->{title_font_button}->set_show_style(TRUE);
	$tit->attach_defaults(&_align_left($self->{title_font_button}),
		1, 2, 0, 1);

	$lab = &_new_label('_Director color:');
	$tit->attach_defaults($lab, 0, 1, 1, 2);
	$self->{director_color_button} = Gtk2::ColorButton->new;
	$lab->set_mnemonic_widget($self->{director_color_button});
	$self->{director_color_button}->set_title('Pick director color');
	$tit->attach_defaults(&_align_left($self->{director_color_button}),
		1, 2, 1, 2);

	$lab = &_new_label('Foc_used color:');
	$tit->attach_defaults($lab, 0, 1, 2, 3);
	$self->{focused_color_button} = Gtk2::ColorButton->new;
	$lab->set_mnemonic_widget($self->{focused_color_button});
	$self->{focused_color_button}->set_title('Pick focused color');
	$tit->attach_defaults(&_align_left($self->{focused_color_button}),
		1, 2, 2, 3);

	$lab = &_new_label('_Free color:');
	$tit->attach_defaults($lab, 0, 1, 3, 4);
	$self->{free_color_button} = Gtk2::ColorButton->new;
	$lab->set_mnemonic_widget($self->{free_color_button});
	$self->{free_color_button}->set_title('Pick free color');
	$tit->attach_defaults(&_align_left($self->{free_color_button}),
		1, 2, 3, 4);

	$self->{show_close_button} = Gtk2::CheckButton->new(
		'Show cl_ose button');
	$tit->attach_defaults($self->{show_close_button}, 0, 2, 4, 5);

	$self->{show_single_role_button} = Gtk2::CheckButton->new(
		'Show single _role button');
	$tit->attach_defaults($self->{show_single_role_button}, 0, 2, 5, 6);

	return &_align_left($tit);
}

sub _new_highlighting {
	my $self = shift;
	my $hi = Gtk2::Table->new(3, 1);
	$hi->set_col_spacings(6);
	$hi->set_row_spacings(6);

	$self->{highlight_foreground_combo} = Gtk2::ComboBox->new_text;
	map { $self->{highlight_foreground_combo}->append_text($_) } @color;
	$self->{highlight_foreground_combo}->signal_connect_swapped(
		changed => \&_changed_highlight_foreground, $self);
	$hi->attach_defaults($self->{highlight_foreground_combo}, 0, 1, 0, 1);

	$hi->attach_defaults(&_new_label('on'), 1, 2, 0, 1);

	$self->{highlight_background_combo} = Gtk2::ComboBox->new_text;
	map { $self->{highlight_background_combo}->append_text($_) } @color;
	$self->{highlight_background_combo}->signal_connect_swapped(
		changed => \&_changed_highlight_background, $self);
	$hi->attach_defaults($self->{highlight_background_combo}, 2, 3, 0, 1);

	return &_align_left($hi);
}

sub _align_left {
	my $a = Gtk2::Alignment->new(0, 0.5, 0, 1.0);
	$a->add($_[0]);
	return $a;
}

1;
