##
## Copyright (c) 2005-2007 by Rafael D'Halleweyn
##
## Licensed under Version 2 of the GPL
##

package ClusTerm::Term;

use strict;
use warnings;
use Exporter 'import';
use Carp;
use Glib qw(TRUE FALSE);
use Gnome2;
use Gnome2::Vte;
use ClusTerm::Utils qw(color_gdk color_rgb color_string color_best_contrast);
use Term::Cap;

our @EXPORT_OK = qw(ROLE_ACTOR ROLE_DIRECTOR ROLE_FREE ROLE_DISCONNECT);
our %EXPORT_TAGS = (
	roles => [qw(ROLE_ACTOR ROLE_DIRECTOR ROLE_FREE)],
);

use constant {
	ROLE_ACTOR => 0,
	ROLE_FREE => 1,
	ROLE_DIRECTOR => 2,
	ROLE_DISCONNECT => 3,
};
my @roles = qw(Actor Free Director);
my @role_stock = qw(gtk-media-play-ltr gtk-media-stop gtk-media-record
	gtk-disconnect);

my @referenced_methods;
BEGIN {
# referenced methods from Gnome2::Vte::Terminal
# needs to be in BEGIN, because the 'use' where it is used also is BEGIN!
@referenced_methods = qw(get_adjustment get_allow_bold set_allow_bold
 get_audible_bell set_audible_bell set_background_saturation
 set_background_tint_color set_background_transparent set_backspace_binding
 get_char_ascent get_char_descent get_char_height get_char_width
 set_color_background set_color_bold set_color_cursor set_color_dim
 set_color_foreground set_color_highlight set_colors get_column_count
 copy_clipboard copy_primary set_cursor_blinks set_default_colors
 get_default_emulation set_delete_binding get_emulation set_emulation
 get_encoding set_encoding feed feed_child set_font_from_string_full
 set_font_from_string set_font_full get_font set_font 
 get_has_selection get_icon_title is_word_char match_add match_check
 match_clear_all match_remove match_set_cursor match_set_cursor_type
 get_mouse_autohide set_mouse_autohide paste_clipboard paste_primary
 reset get_row_count set_scroll_background set_scroll_on_keystroke
 set_scroll_on_output set_scrollback_lines set_size get_status_line
 get_text get_text_include_trailing_spaces get_text_range get_using_xft
 get_visible_bell set_visible_bell get_window_title set_word_chars
 set_background_image set_background_image_file get_cursor_position);
}

my $term_font = 'Andale Mono 7';
my @term_colors = qw(f5f5f5 000000
	000000 aa0000 00aa00 aa5500 0000aa aa00aa 00aaaa e1e1e1
	555555 ff5555 55ff55 ffff55 5555ff ff55ff 55ffff ffffff);
my $gnome_term_gconf_dir = "/apps/gnome-terminal/profiles/";

use ClusTerm::GlibSubclass Gtk2::EventBox::,
	references => {
		term => [ Gnome2::Vte::Terminal::,
			'commit',
			'contents_changed',
			@referenced_methods,
		],
		scrollbar => [ Gtk2::VScrollbar::,
			[ get_value => 'get_scroll' ],
			[ set_value => 'set_scroll' ],
		],
	},
	signals => {
		role_change => { },
		context_menu => { param_types =>
			[ 'Gtk2::Gdk::Event', 'Gtk2::Widget' ]},
		close => { param_types => [ 'Glib::String', 'Glib::Int' ] },
	},
	properties => [
		{
			pspec => Glib::ParamSpec->string(
				'profile', 'Profile',
				'Gnome Terminal Profile',
				'Default',
				[qw/readable writable/]),
			set => \&set_profile,
		},
		{
			pspec => Glib::ParamSpec->string(
				'title-font-name', 'Title font name',
				'Font user for the title',
				'Sans Bold 10',
				[qw/readable writable/]),
			set => \&set_title_font,
		},
		{
			pspec => Glib::ParamSpec->boxed(
				'director-color', 'Director Color',
				'Title background color when role is '
					.'director and focused',
				'Gtk2::Gdk::Color',
				[qw/readable writable/]),
			set => \&set_director_color,
		},
		{
			pspec => Glib::ParamSpec->boxed(
				'focused-color', 'Focused Color',
				'Title background color when focused',
				'Gtk2::Gdk::Color',
				[qw/readable writable/]),
			set => \&set_focused_color,
		},
		{
			pspec => Glib::ParamSpec->boxed(
				'free-color', 'Free Color',
				'Title background color when role is '
					.'free and unfucosed',
				'Gtk2::Gdk::Color',
				[qw/readable writable/]),
			set => \&set_free_color,
		},
		{
			pspec => Glib::ParamSpec->boxed(
				'unfocused-color', 'Unfocused Color',
				'Title background color when unfocused',
				'Gtk2::Gdk::Color',
				[qw/readable writable/]),
			set => \&set_unfocused_color,
		},
		{
			pspec => Glib::ParamSpec->IV(
				'role', 'Role',
				'Role of terminal: actor, director or free',
				ROLE_ACTOR, ROLE_DIRECTOR, ROLE_ACTOR,
				[qw/readable writable/]),
			set => \&set_role,
		},
		{
			pspec => Glib::ParamSpec->boolean(
				"showing_context_menu", "",
				"If the terminal is showing a context menu",
				FALSE,
				[qw/readable writable/]),
			set => \&set_showing_context_menu,
		},
		{
			pspec => Glib::ParamSpec->boolean(
				"has_close_button", "",
				"If the terminal is showing a context menu",
				FALSE,
				[qw/readable writable/]),
			set => \&set_has_close_button,
		},
		{
			pspec => Glib::ParamSpec->boolean(
				"has_single_role_button", "",
				"If the terminal is showing a context menu",
				FALSE,
				[qw/readable writable/]),
			set => \&set_has_single_role_button,
		},
		{
			pspec => Glib::ParamSpec->boolean(
				"close_on_disconnect", "",
				"If the terminal closes when the client "
					."disconnects",
				TRUE,
				[qw/readable writable/]),
			#set => \&set_close_on_disconnect,
		},
	];

sub new {
	my ($class, $name, $role, $userdata) = @_;
	my $self = Glib::Object::new(__PACKAGE__);

	$self->{name} = $name;
	$self->{title} = $name;
	$self->{role} = defined $role ? $role : ROLE_ACTOR;
	$self->{focused} = FALSE;
	$self->{userdata} = $userdata;
	$self->{start_time} = time;

	# we want to filter events before to go to the term
	$self->set_visible_window(TRUE);
	$self->add_events([ qw(focus_change_mask button_press_mask) ]);
	$self->signal_connect_swapped(
		button_press_event => \&_button_press, $self);
	$self->signal_connect_swapped(
		focus_in_event => \&_focus_in, $self);
	$self->can_focus(TRUE);

	# two rows: label and vte
	$self->{vbox} = Gtk2::VBox->new(FALSE, 0);
	$self->add($self->{vbox});

	$self->{hbox_term} = Gtk2::HBox->new(FALSE, 0);
	$self->{hbox_label} = Gtk2::HBox->new(FALSE, 0);
	$self->{vbox}->pack_start($self->{hbox_label}, FALSE, FALSE, 0);
	$self->{vbox}->pack_start($self->{hbox_term}, TRUE, TRUE, 0);

	# label (vte title)
	$self->_create_label;

	# icons for record/play/pause
	$self->{icon_theme} = Gtk2::IconTheme->get_default;
	$self->_create_icons;

	# role buttons
	$self->_create_role_buttons;

	# close button?
	$self->_create_close_button if $self->{has_close_button};

	# vte
	$self->_create_term;

	# scrollbar
	$self->{scrollbar} = Gtk2::VScrollbar->new;
	$self->{scrollbar}->set_adjustment($self->{term}->get_adjustment);
	$self->{hbox_term}->pack_start($self->{scrollbar}, FALSE, FALSE, 0);

	$self->{finished} = 1;
	$self->set(
		director_color => &color_gdk('#c90000'),
		focused_color => &color_gdk('#0076c9'),
		free_color => &color_gdk('#e6d6a1'),
		unfocused_color => &color_gdk('#bfbfbf')
	);
	$self->_set_back;

	# output functions
	$self->{out} = "";
	$self->{color_hash} = undef;
	$self->{tcap} = undef;

	$self->{gconf} = Gnome2::GConf::Client->get_default;
	$self->{gconf}->signal_connect_swapped(
		value_changed => \&_gconf_gnome_term_profile_changed, $self);

	return $self->subclass_connect;
}

sub age {
	return time - $_[0]->{start_time};
}

sub get_name {
	$_[0]->{name};
}
sub set_arg {
	$_[0]->{arg} = $_[1];
}

sub get_arg {
	$_[0]->{arg};
}


sub _create_term {
	my ($self) = @_;

	$self->{term} = Gnome2::Vte::Terminal->new;

	$self->{term}->add_events([ qw(focus_change_mask button_press_mask) ]);
	$self->{term}->signal_connect_swapped(
		focus_in_event => \&_focus_in, $self);
	$self->{term}->signal_connect_swapped(
		focus_out_event => \&_focus_out, $self);
	$self->{term}->signal_connect_swapped(
		window_title_changed => \&_change_window_title, $self);
	$self->{term}->signal_connect_swapped(
		button_press_event => \&_button_press, $self);
	$self->{term}->signal_connect_swapped(
		key_press_event => \&_key_press, $self);
	$self->{term}->signal_connect_swapped(
		popup_menu => \&_popup_menu, $self);
	$self->{term}->signal_connect_swapped(
		child_exited => \&_child_exited, $self);

	$self->{hbox_term}->pack_start($self->{term}, TRUE, TRUE, 0);
}

sub _create_label {
	my ($self) = @_;

	$self->{label} = Gtk2::Label->new();
	$self->{label}->set_use_markup(TRUE);
	$self->{label}->set_justify("left");
	$self->{label}->set_single_line_mode(TRUE);
	$self->{label}->can_focus(TRUE);
	$self->{label}->signal_connect_swapped(focus => \&_focus, $self);
	$self->{label}->set_ellipsize("end");
	$self->{label}->set_alignment(0.0, 0.0);
	$self->{label}->set_padding(2, 0);
	$self->{hbox_label}->pack_start($self->{label}, TRUE, TRUE, 0);

	# measure the size of the label
	$self->{title_font_name} ||= 'Sans Bold 10';
	my $layout = $self->{label}->create_pango_layout("g\@F");
	my $fd = Gtk2::Pango::FontDescription->from_string(
		$self->{title_font_name});
	$layout->set_font_description($fd);
	$self->{label_height} = ($layout->get_pixel_size)[1];
	$self->{label_height} = 16;

	$self->_set_title($self->{name});
}

sub _create_icons {
	my ($self) = @_;

	# delete any icons we may have
	$self->{role_image} = [ ];

	# the size of the icon should be 90% of the font pixel size,
	# but never less than 16
	my $size = $self->{label_height};
	$size = int($size * .75);
	$size = 16 unless ($size > 16);

	for (@role_stock) {
		my $icon;
		eval {
			$icon = $self->{icon_theme}->load_icon($_, $size, [ ]);
		};
		push @{$self->{role_image}}, 
			Gtk2::Image->new_from_pixbuf($icon);
	}

	my $icon = $self->{icon_theme}->load_icon("gtk-close", $size, [ ]);
	$self->{close_icon} = Gtk2::Image->new_from_pixbuf($icon);

	# we don't always get what we ask for
	$size = $self->{role_image}[0]
		? $self->{role_image}[0]->size_request->height
		: 8;

	# the button size is the font pixel size, but never less
	# than 8 pixels bigger than the icon
	$self->{button_size} = $self->{label_height};
	$self->{button_size} = $size + 8
		if ($self->{button_size} < $size + 8);
}

sub _create_role_buttons {
	my ($self) = @_;

	delete $self->{role_button};
	if ($self->{has_single_role_button}) {
		$self->{role_button} = Gtk2::Button->new;
		$self->{role_button}->add($self->{role_image}[$self->{role}])
			if $self->{role_image}[$self->{role}];
		$self->{role_button}->set_size_request(
			$self->{button_size}, $self->{button_size});
		$self->{role_button}->can_focus(FALSE);
		$self->{role_button}->add_events('button_press_mask');
		$self->{role_button}->signal_connect_swapped(
			button_press_event => \&_role_button, $self);
		$self->{hbox_label}->pack_start($self->{role_button},
			FALSE, FALSE, 0);
		$self->{hbox_label}->reorder_child($self->{role_button}, 1);
	} else {
		for (ROLE_FREE, ROLE_ACTOR, ROLE_DIRECTOR) {
			$self->{role_button}[$_] = Gtk2::ToggleButton->new;
			$self->{role_button}[$_]->add(
				$self->{role_image}[$_])
				if $self->{role_image}[$_];
			$self->{role_button}[$_]->set_size_request(
				$self->{button_size}, $self->{button_size});
			$self->{role_button}[$_]->can_focus(FALSE);
			$self->{role_button}[$_]->set_active(
				$_ == $self->{role});
			$self->{role_button}[$_]->add_events(
				'button_press_mask');
			$self->{role_button}[$_]->signal_connect_swapped(
				button_press_event => \&_role_toggle, $self);
			$self->{hbox_label}->pack_start(
				$self->{role_button}[$_], FALSE, FALSE, 0);

			# adding in reverse order!!
			$self->{hbox_label}->reorder_child(
				$self->{role_button}[$_], 1);
		}
	}
}

sub _create_close_button {
	my ($self) = @_;

	$self->{close_button} = Gtk2::Button->new;
	$self->{close_button}->add($self->{close_icon});
	$self->{close_button}->set_size_request(
		$self->{button_size}, $self->{button_size});
	$self->{close_button}->can_focus(FALSE);
	$self->{close_button}->signal_connect_swapped(
		clicked => sub {
			$_[0]->signal_emit('close', 'button', 0)
		}, $self);
	$self->{hbox_label}->pack_start($self->{close_button},
		FALSE, FALSE, 0);
	return $self->{close_button};
}

sub get_padding {
	my $self = shift;
	my ($xpad, $ypad) = $self->{term}->get_padding;

	# add label height and scrollbar width
	my $scroll_req = $self->{scrollbar}->size_request;
	$xpad += $scroll_req->width;
	$ypad += $self->{button_size};
	return ($xpad, $ypad);
}

sub _change_window_title {
	my ($self) = @_;
	$self->_set_title($self->{term}->get_window_title);
}

sub _set_title {
	my ($self, $title) = @_;
	
	if ($title) {
		$self->{title} = $title;
	} else {
		$title = $self->{title};
	}

	my $fg = $self->{fg_string};
	$title = "<span font_desc=\"$self->{title_font_name}\""
			.($fg ? " foreground=\"$fg\"" : "").">"
		.$title
		."</span>";
	$self->{label}->set_label($title);
}

sub _set_back {
	my ($self) = @_;

	return unless $self->{finished};

	my $bg_name;
	if ($self->{focused} || $self->{showing_context_menu}) {
		$bg_name = ($self->{role} == ROLE_DIRECTOR)
			? 'director_color' : 'focused_color';
	} else {
		$bg_name = ($self->{role} == ROLE_DIRECTOR)
			? 'unfocused_director_color'
			: ($self->{role} == ROLE_FREE)
			? 'free_color'
			: 'unfocused_color';
	}

	my $bg = $self->{$bg_name};
	if ($bg) {
		$self->modify_bg('normal', $bg);

		my $fg;
		if ($self->{focused}) {
			$fg = &color_best_contrast($bg,
				[ 0, 0, 0], [ 255, 255, 255 ]);
		} else {
			$fg = &color_best_contrast($bg,
				[ 51, 51, 51 ], [ 231, 231, 231 ]);
		}
		$self->{fg_string} = &color_string($fg);
	}

	my $align = 0.0;
	$align = ($self->{button_size} - $self->{label_height})
			* 1.3 / $self->{label_height}
		unless $self->{button_size} == $self->{label_height};
	$align = 1.0 if $align > 1.0;
	$self->{label}->set_alignment(0.0, $align);

	$self->_set_title;
}

sub _role_button {
	my ($self, $button, $obj) = @_;

	return FALSE unless $button->button == 1;
	my $type = $button->type;
	if ($type eq '2button-press') {
		$self->set_role(ROLE_DIRECTOR);
	} elsif ($type eq 'button-press') {
		$self->set_role($self->{role} == ROLE_FREE
			? ROLE_ACTOR : ROLE_FREE);
	} else {
		return FALSE;
	}
	$self->signal_emit("role_change");
	return TRUE;
}

sub _role_toggle {
	my ($self, $button, $obj) = @_;
	return FALSE
		unless $button->button == 1 && $button->type eq 'button-press';

	my $maxrole = $#{$self->{role_button}};
	my ($new_role) = grep { $self->{role_button}[$_] == $obj }
		(0..$maxrole);
	if ($new_role != $self->{role}) {
		$self->set_role($new_role);
	} else {
		$self->set_role($self->{role} == ROLE_FREE
			? ROLE_ACTOR : ROLE_FREE);
	}
	$self->grab_focus if ($self->{role} == ROLE_DIRECTOR);
	$self->signal_emit("role_change");

	return TRUE;
}

sub set_role {
	my ($self, $newrole) = @_;

	if ($self->{has_single_role_button}) {
		$self->{role_button}->remove(
			$self->{role_image}[$self->{role}])
			if $self->{role_image}[$self->{role}];
		$self->{role_button}->add($self->{role_image}[$newrole])
			if $self->{role_image}[$newrole];
		$self->{role_button}->show_all;
		$self->{term}->grab_focus if ($newrole == ROLE_DIRECTOR);
	} else {
		$self->{role_button}[$self->{role}]->set_active(FALSE);
		$self->{role_button}[$newrole]->set_active(TRUE);
	}

	$self->{role} = $newrole;
	$self->_set_back;
}

sub set_showing_context_menu {
	my ($self, $showing) = @_;
	$self->{showing_context_menu} = $showing;
	$self->_set_back;
}

sub set_has_single_role_button {
	my ($self, $single) = @_;
	# unchanged state
	return if (!! $self->{has_single_role_button}) == $single;

	$self->_redo_role_button($single);
	$self->{hbox_label}->show_all;
}

sub _redo_role_button {
	my ($self, $single) = @_;
	$single = $self->{has_single_role_button} unless defined $single;

	# remove old button(s)
	if ($self->{has_single_role_button}) {
		# remove single button
		# have to remove the icon first, otherwise the 
		# remove will cascade destroy
		my ($icon) = $self->{role_button}->get_children;
		$self->{role_button}->remove($icon) if $icon;
		$self->{hbox_label}->remove($self->{role_button});
	} else {
		# remove three buttons
		for (ROLE_ACTOR, ROLE_FREE, ROLE_DIRECTOR) {
			my ($icon) = $self->{role_button}[$_]->get_children;
			$self->{role_button}[$_]->remove($icon)
				if $icon;
			$self->{hbox_label}->remove($self->{role_button}[$_]);
		}
	}

	# add new buttons
	$self->{has_single_role_button} = $single;
	$self->_create_role_buttons;
}

sub set_has_close_button {
	my ($self, $close) = @_;
	# unchanged state
	return if (!! $self->{has_close_button}) == $close;

	$self->_redo_close_button($close);
	$self->{hbox_label}->show_all;
}

sub _redo_close_button {
	my ($self, $close) = @_;
	$close = $self->{has_close_button} unless defined $close;

	if ($self->{has_close_button}) {
		# remove the close button
		# have to remove the icon first, otherwise the 
		# remove will cascade destroy
		my ($icon) = $self->{close_button}->get_children;
		$self->{close_button}->remove($icon);
		$self->{hbox_label}->remove($self->{close_button});
		delete $self->{close_button};
	}
	$self->{has_close_button} = $close;
	if ($self->{has_close_button}) {
		# show the close button
		$self->_create_close_button;
	}
}

my %colors = (Black => 0, Red => 1, Green => 2, Yellow => 3, Blue => 4,
	Magenta => 5, Cyan => 6, White => 7);

sub output {
	my ($self, $text, $attrs, $override) = @_;

	$self->_get_tcap();
	$self->_get_color_hash();
	my $ansi = $self->{term_emul} =~ /xterm|gnome|ansi/;
	my $tcap = $self->{tcap};

	my $last_markup = undef;
	my $len = length($text);
	my $t = "";
	for(my $i = 0; $i < $len; $i++) {
		my %attr = (
			$attrs && $i < @$attrs ? %{$attrs->[$i]} : ( ),
			$override ? %$override : ( ),
		);

		my $h;
		my $fg = exists $attr{fg}
			? $attr{fg}
			: ! exists $attr{fore}
			? -2
			: do { $h = $attr{fore}->hash;
				exists $self->{color_hash}{$h} }
			? $self->{color_hash}{$h}: -2;
		$fg = $colors{$fg} if exists $colors{$fg};
		my $bg = exists $attr{bg}
			? $attr{bg}
			: ! exists $attr{back}
			? -1
			: do { $h = $attr{back}->hash;
				exists $self->{color_hash}{$h} }
			? $self->{color_hash}{$h}: -1;
		$bg = $colors{$bg} if exists $colors{$bg};

		my $markup = "";
		if ($ansi) {
			my @a = (0);

			if ($fg == -1 || $bg == -2) {
				($fg, $bg) = ($bg, $fg);
				push @a, 7;
			}
			if ($fg >= 0 && $fg <= 7) {
				push @a, 30 + $fg;
			} elsif ($fg >= 8 && $fg <= 15) {
				push @a, 2, 30 + $fg - 8;
			}
			if ($bg >= 0 && $bg <= 7) {
				push @a, 40 + $bg;
			} elsif ($bg >= 8 && $bg <= 15) {
				push @a, 2, 30 + $bg - 8;
			}
			# bold and italic are not in markup...
			push @a, 1 if $attr{bold};
			push @a, 3 if $attr{italic};
			push @a, 4 if $attr{underline};
			push @a, 9 if $attr{strikethrough};

			$markup = "\e[".join(";", @a)."m" if @a > 1;
			unless ($last_markup && $last_markup eq $markup) {
				$t .= $markup;
				$last_markup = $markup;
			}
		} else {
			$markup .= $tcap->Tputs('mr') || ""
				if $bg != -1 || $fg != -2;
			$markup .= $tcap->Tputs('ue') || ""
				if $attr{underline};
			$markup .= $tcap->Tputs('ZH') || ""
				if $attr{italic};
			$markup .= $tcap->Tputs('md') || ""
				if $attr{bold};
			unless ($last_markup && $last_markup eq $markup) {
				$t .= $tcap->Tputs('me') . $markup;
				$last_markup = $markup;
			}
		}

		my $char = substr($text, $i, 1);
		$t .= $char eq "\n" ? "\r\n" : $char;
	}
	$t .= $ansi ? "\e[0m" : $tcap->Tputs('me');

	$self->{out} .= $t;
}

sub output_set_cursor {
	my ($self, $col, $row) = @_;
	$self->_get_tcap();
	$self->{out} .= $self->{tcap}->Tgoto('cm', $col, $row);
}

sub output_flush {
	my ($self) = @_;

	$self->feed($self->{out});

	$self->{out} = "";
	$self->{color_hash} = undef;
	$self->{tcap} = undef;
}

sub set_disconnected {
	my ($self) = @_;

	$self->sensitive(0);
	#$self->{term}->sensitive(0);
	if ($self->{has_single_role_button}) {
		$self->{role_button}->sensitive(0);
	} else {
		for (ROLE_FREE, ROLE_ACTOR, ROLE_DIRECTOR) {
			$self->{role_button}[$_]->sensitive(0);
		}
	}
	$self->{label}->sensitive(0);
	$self->set_role(ROLE_FREE);
	$self->_set_title("Closed: " . $self->{title});

	$self->{reconnect} = Gtk2::Button->new;
	$self->{reconnect}->add($self->{role_image}[ROLE_DISCONNECT]);
	$self->{reconnect}->set_size_request(
		$self->{button_size}, $self->{button_size});
	$self->{reconnect}->can_focus(FALSE);
	$self->{reconnect}->add_events('button_press_mask');
	$self->{reconnect}->signal_connect_swapped(
		button_press_event => \&_reconnect_button, $self);
	$self->{hbox_label}->pack_start($self->{reconnect},
		FALSE, FALSE, 0);
	$self->{hbox_label}->reorder_child($self->{reconnect}, 1);
	$self->{hbox_label}->show_all();
}

sub _reconnect_button {
	my ($self, $button, $obj) = @_;

	$self->sensitive(1);
	#$self->{term}->sensitive(1);
	if ($self->{has_single_role_button}) {
		$self->{role_button}->sensitive(1);
	} else {
		for (ROLE_FREE, ROLE_ACTOR, ROLE_DIRECTOR) {
			$self->{role_button}[$_]->sensitive(1);
		}
	}
	$self->{label}->sensitive(1);
	$self->set_role(ROLE_ACTOR);
	$self->_set_title($self->{name});

	$self->{hbox_label}->remove($self->{reconnect});
	$self->{reconnect}->remove($self->{role_image}[ROLE_DISCONNECT]);
	delete $self->{reconnect};
	$self->{hbox_label}->show_all();

	$self->fork_command(@{$self->{reconnect_args}});
}

sub _get_tcap {
	return if $_[0]->{tcap};
	$_[0]->{term_emul} = $_[0]->get_emulation();
	$_[0]->{tcap} = Term::Cap->Tgetent({ TERM => $_[0]->{term_emul} });
	unless ($_[0]->{tcap}) {
		carp("Uknown terminal type $_[0]->{term_emul}, "
			."falling back to xterm!");
		$_[0]->{term_emul} = 'xterm';
		$_[0]->{tcap} = Term::Cap->Tgetent(
			{ TERM => $_[0]->{term_emul} });
	}
}

sub _get_color_hash {
	return if $_[0]->{color_hash};
	for(my $i = 0; $i < @{$_[0]->{colors}}; $i++) {
		$_[0]->{color_hash}{$_[0]->{colors}[$i]->hash} = $i - 2;
	}
}

sub set_profile {
	my ($self, $profile) = @_;
	$self->{gconf}->remove_dir($gnome_term_gconf_dir.$self->{profile})
		if $self->{profile};
	$self->{profile} = $profile;
	my $dir = $gnome_term_gconf_dir.$profile;

	unless ($self->{gconf}->dir_exists($dir)) {
		$self->set_size(10, 6);
		$self->{font} = $term_font;
		$self->set_font_from_string($self->{font});
		$self->set_audible_bell(FALSE);
		$self->set_visible_bell(TRUE);
		$self->set_allow_bold(TRUE);
		$self->set_scroll_on_keystroke(TRUE);
		$self->set_scrollback_lines(512);
		$self->set_word_chars('-A-Za-z0-9./?%&#_+');

		unless (ref $term_colors[0]) {
			map { $_ = Gtk2::Gdk::Color->parse('#'.$_) }
				@term_colors;
		}
		$self->set_color_foreground($term_colors[0]);
		$self->set_color_background($term_colors[1]);
		$self->set_colors($term_colors[0], $term_colors[1],
			[ @term_colors[2..$#term_colors] ]);
		$self->{colors} = \@term_colors;
		return;
	}

	$self->{gconf}->add_dir($dir, 'preload-onelevel');
	$self->_set_profile($dir);
}

sub _gconf_gnome_term_profile_changed {
	my ($self, $key, $gptr, $client) = @_;

	return unless $key =~ m{^\Q$gnome_term_gconf_dir\E};
	$self->_set_profile();
}

sub _set_profile {
	my ($self) = @_;
	my $val;

	my $dir = $gnome_term_gconf_dir.$self->{profile};

	#$self->set_($val)
	#if $val = $self->{gconf}->get_bool("$dir/use_theme_colors", FALSE);

	my $fg = $self->{gconf}->get_string("$dir/foreground_color", FALSE);
	my $bg = $self->{gconf}->get_string("$dir/background_color", FALSE);
	$fg = Gtk2::Gdk::Color->parse($fg || '#ffffff');
	$bg = Gtk2::Gdk::Color->parse($bg || '#000000');
	my $palette = $self->{gconf}->get_string("$dir/palette", FALSE);
	if ($palette) {
		my @pal = map { Gtk2::Gdk::Color->parse($_) }
			split /:/, $palette;
		$self->set_colors($fg, $bg, \@pal);
		$self->{colors} = [ $fg, $bg, @pal ];
	}
	$self->set_color_foreground($fg);
	$self->set_color_background($bg);
	$self->set_background_tint_color($bg);

	my $back = $self->{gconf}->get_string("$dir/background_type", FALSE);
	if (! defined $back) {
	} elsif ($back eq 'solid') {
		$self->set_background_transparent(0);
		eval { $self->set_background_image(undef) };
	} elsif ($back eq 'image') {
		$self->set_background_transparent(1);
		$self->set_background_image_file($val)
			if $val = $self->{gconf}->get_string(
				"$dir/background_image", FALSE);
		$self->set_background_saturation(1 - $val)
			if $val = $self->{gconf}->get_float(
				"$dir/background_darkness", FALSE);
	} elsif ($back eq 'transparent') {
		$self->set_background_transparent(1);
		eval { $self->set_background_image(undef) };
		$self->set_background_saturation(1 - $val)
			if $val = $self->{gconf}->get_float(
				"$dir/background_darkness", FALSE);
	}

	#$self->set_($val)
	#if $val = $self->{gconf}->get_bool("$dir/use_system_font", FALSE);
	$self->{font} = $self->{gconf}->get_string("$dir/font", FALSE);
	$self->{font} = $term_font
		unless $self->{font}
		&& ! $self->{gconf}->get_bool("$dir/use_system_font", FALSE);
	$self->set_font_from_string($self->{font});

	$self->set_allow_bold($val)
		if defined ($val = $self->{gconf}->get_bool("$dir/allow_bold",
		FALSE));
	$self->set_backspace_binding(&_to_erase_binding($val))
		if $val = $self->{gconf}->get_string("$dir/backspace_binding",
		FALSE);
	$self->set_cursor_blinks($val)
		if defined ($val = $self->{gconf}->get_bool(
		"$dir/cursor_blink", FALSE));
	$self->set_delete_binding(&_to_erase_binding($val))
		if $val = $self->{gconf}->get_string("$dir/delete_binding",
		FALSE);
	$self->set_scroll_background($val)
		if defined ($val = $self->{gconf}->get_bool(
		"$dir/scroll_background", FALSE));
	$self->set_scrollback_lines($val)
		if $val = $self->{gconf}->get_int("$dir/scrollback_lines",
		FALSE);
	$self->set_scroll_on_keystroke($val)
		if defined ($val = $self->{gconf}->get_bool(
		"$dir/scroll_on_keystroke", FALSE));
	$self->set_scroll_on_output($val)
		if defined ($val = $self->{gconf}->get_bool(
		"$dir/scroll_on_output", FALSE));
	$self->set_audible_bell(! $val)
		if defined ($val = $self->{gconf}->get_bool(
		"$dir/silent_bell", FALSE));
	$self->set_word_chars($val)
		if $val = $self->{gconf}->get_string("$dir/word_chars", FALSE);
}

sub set_zoom {
	my ($self, $zoom) = @_;

	return unless $self->{font} =~ /^(.+)\s+(\d+)$/;
	my ($name, $size) = ($1, $2);

	my $new_size = int($size * $zoom + .5);
	if ($zoom != 1 && $new_size <= 10) {
		$new_size = $size + ($zoom - 1) * 5;
	}

	$new_size = 4 if $new_size < 4;

	$self->set_font_from_string("$name $new_size");
}

sub _to_erase_binding {
	if ($_[0] eq 'ascii-del') {
		return 'ascii-delete';
	} elsif ($_[0] eq 'escape-sequence') {
		return 'delete-sequence';
	} elsif ($_[0] eq 'control-h') {
		return 'ascii-backspace';
	} else {
		return 'auto';
	}
}

sub set_title_font {
	my ($self, $f) = @_;
	return if $self->{title_font_name} eq $f;

	$self->{title_font_name} = $f;

	my $layout = $self->{label}->create_pango_layout("g\@F");
	my $fd = Gtk2::Pango::FontDescription->from_string(
		$self->{title_font_name});
	$layout->set_font_description($fd);
	$self->{label_height} = ($layout->get_pixel_size)[1];

	$self->_create_icons;
	$self->_redo_role_button;
	$self->_redo_close_button;
	$self->_set_back;

	$self->show_all;
}

sub set_director_color {
	my ($self, $c) = @_;
	$self->{director_color} = &color_gdk($c);
	$self->{unfocused_director_color} = &color_gdk(
		map { 134 + sqrt(sqrt($_)) * 25 } &color_rgb($c));
	$self->_set_back if $self->{role} == ROLE_DIRECTOR;
}

sub set_focused_color {
	my ($self, $c) = @_;
	$self->{focused_color} = &color_gdk($c);
	$self->_set_back if $self->{role} != ROLE_DIRECTOR
		&& $self->{focused};
}

sub set_free_color {
	my ($self, $c) = @_;
	$self->{free_color} = &color_gdk($c);
	$self->_set_back if $self->{role} == ROLE_FREE
		&& ! $self->{focused};
}

sub set_unfocused_color {
	my ($self, $c) = @_;
	$self->{unfocused_color} = &color_gdk($c);
	$self->_set_back if ! $self->{focused}
		&& $self->{role} != ROLE_DIRECTOR;
}

sub get_active_text {
	my ($self) = @_;

	my $oldscroll = $self->get_scroll();
	# scroll to the end, to make sure we get the 'active' text
	$self->set_scroll(~0);
	#Glib::MainContext->default()->iteration(TRUE);

	#my $start_row = $self->get_scroll();
	#my ($text, $attrs) = $self->get_text_include_trailing_spaces(sub { 1 });
	my ($text, $attrs) = $self->get_text();
	#my ($text, $attrs) = $self->get_text_range(
	#	$start_row, 0,
	#	$start_row + $self->get_column_count() - 1,
	#	$self->get_row_count() - 1,
	#	sub { 1 });

	# scroll back
	$self->set_scroll($oldscroll);

	return ($text, $attrs);
}

sub _focus_in {
	my ($self, $event, $obj) = @_;
	my $old_focus = $self->{focused};
	$self->{focused} = TRUE;
	$self->_set_back if ! $old_focus;
	$self->{term}->grab_focus;

	$self->signal_emit('focus_in_event', $event)
		unless ($self == $obj);

	return FALSE;
}

sub _focus_out {
	my ($self) = @_;
	$self->{focused} = FALSE;

	# if a context menu is active, we still want to show us as
	# focused, even though the menu really has the focus
	$self->_set_back unless $self->{context_menu};
	delete $self->{context_menu};

	return FALSE;
}

sub _button_press {
	my ($self, $button, $obj) = @_;

	if ($button->button == 3) {
		$self->{context_menu} = TRUE;
		$self->signal_emit("context_menu", $button, $obj);
		return TRUE;
	}

	my $type = $button->type;
	if ($obj == $self) {
		$self->grab_focus;
		if ($type eq '2button-press') {
			$self->set_role(ROLE_DIRECTOR);
			$self->signal_emit("role_change");
		}
		return TRUE;
	}
	return FALSE;
}

sub _popup_menu {
	my ($self) = @_;

	$self->{context_menu} = TRUE;
	$self->signal_emit("context_menu", undef, undef);
	return TRUE;
}

sub _child_exited {
	my ($self, $rc) = @_;

	if ($self->{close_on_disconnect}) {
		$self->signal_emit('close', 'child-exit', $rc || 0);
	} else {
		$self->set_disconnected();
	}
}

sub _key_press {
	my ($self, $key, $obj) = @_;

	# propogate the event to our own object
	return $self->signal_emit('key_press_event', $key);
}

my $clusterm_helper;
my ($helper_read, $helper_write);
my %term_cache;
sub fork_command {
	my ($self, $cmd, $arg, @other) = @_;

	$self->{reconnect_args} = [ $cmd, [ @$arg ], @other ];

#	if (1) {
#		# this doesn't work with older versions of Gnome2::Vte??
#		my $pid = $self->{term}->fork_command($cmd, $arg, @other);
#		$term_cache{$pid} = $self;
#		return;
#	}

	if (@$arg) {
		$arg->[0] = $cmd;
	} else {
		push @$arg, $cmd;
	}

	unless ($clusterm_helper) {
		for (@INC) {
			next unless -x "$_/clusterm-helper";
			$clusterm_helper = "$_/clusterm-helper";
			last;
		}
	}

	unless ($helper_read && $helper_write) {
		pipe($helper_read, $helper_write) or
			return 1;

		Glib::IO->add_watch(fileno($helper_read), 'in', \&helper_io);
	}

	return 1 unless $clusterm_helper;
	unshift @$arg, "clusterm-helper", "/proc/$$/fd/".fileno($helper_write);
	my $pid = $self->{term}->fork_command($clusterm_helper, $arg, @other);

	$term_cache{$pid} = $self;
	return $pid;
}

sub helper_io {
	my $buf;
	sysread($helper_read, $buf, 256);

	for my $line (split /\n/, $buf) {
		if ($line =~ /^end (\d+)(?:\s+(\d+))?$/i
			&& exists $term_cache{$1}) {
			my $self = $term_cache{$1};
			$self->_child_exited($2);
		} else {
			carp("Unknown helper communication: $line");
		}
	}

	return TRUE;
}

1;
