##
## Copyright (c) 2005-2007 by Rafael D'Halleweyn
##
## Licensed under Version 2 of the GPL
##

package ClusTerm::Utils;

use strict;
use warnings;
use Exporter 'import';
use Glib qw(TRUE FALSE);
use Gtk2;
use Carp;

our @EXPORT_OK = qw(color_gdk color_rgb color_best_contrast color_string);

sub uiinfo_accelerators {
	my ($uii) = @_;

	return unless (ref $uii && ref $uii eq 'ARRAY');
	for (@$uii) {
		if (exists $_->{accelerator}) {
			my ($k, $m) = Gtk2::Accelerator->parse(
				$_->{accelerator});
			delete $_->{accelerator};
			$_->{accelerator_key} = $k;
			$_->{ac_mods} = $m;
		}
		&uiinfo_accelerators($_->{subtree})
			if (exists $_->{subtree});
	}
}

sub uiinfo_swapped {
	my ($uii, $obj) = @_;

	return unless (ref $uii && ref $uii eq 'ARRAY');
	for(@$uii) {
		&uiinfo_swapped($_->{moreinfo}, $obj)
			if exists $_->{moreinfo}
			&& ref $_->{moreinfo} eq 'ARRAY';
		next unless exists $_->{callback};
		my $cb = $_->{callback};

		$_->{callback} = sub { $cb->($obj, @_) };
	}
}

my $theme;
sub uiinfo_icons {
	my ($uii) = @_;

	$theme ||= Gtk2::IconTheme->get_default;

	return unless (ref $uii && ref $uii eq 'ARRAY');
	for(@$uii) {
		next unless exists $_->{pixmap_type}
			&& $_->{pixmap_type} eq 'icon';
		my $name = $_->{pixmap_info};
		my $icon;

		if ($theme->has_icon($name)) {
			$icon = $theme->lookup_icon($name, 64, [ ]);
		} else {
			print STDERR "No icon named '$name' in theme.\n";
			delete $_->{pixmap_type};
			delete $_->{pixmap_info};
			next;
		}
		if (1 && $icon) {
			$_->{pixmap_type} = 'filename';
			$_->{pixmap_info} = $icon->get_filename;
			# need to convert it to xpm string
		} else {
			delete $_->{pixmap_type};
			delete $_->{pixmap_info};
		}
	}
}

sub uiinfo_store_widgets {
	my ($uii, $shell, $pre, $radio) = @_;
	$pre ||= '';

	my ($menu_last, @menu_last_children);

	for(@$uii) {
		next unless $_->{type};
		if ($_->{type} eq 'radioitems') {
			&uiinfo_store_widgets($_->{moreinfo}, $shell, $pre, 1)
		} elsif ($_->{type} =~ /^subtree/) {
			&uiinfo_store_widgets($_->{subtree}, $shell,
				"$pre$_->{label}/", $radio);
		} elsif ($_->{type} =~ /toggleitem/
			|| ($_->{type} =~ /item/ && $radio)) {
			my $path = $pre.$_->{label};
			my ($menu, $pos) = $shell->find_menu_pos($path);
			if (!$menu_last || $menu_last != $menu) {
				$menu_last = $menu;
				@menu_last_children = $menu->get_children;
			}
			if ($pos > 0 && $pos <= @menu_last_children) {
				push @{$_->{widgets}},
					$menu_last_children[$pos-1];
			} else {
				carp("Did not find a widget at '$path'.");
			}
		} else {
			push @{$_->{widgets}}, $_->{widget};
		}
	}
}

sub color_from {
	return sprintf "#%02x%02x%02x",
		$_[0]->red >> 8, $_[0]->green >> 8, $_[0]->blue >> 8;
}

sub color_rgb {
	if (! defined $_[0]) {
		return (0, 0, 0);
	} elsif (@_ == 3) {
		return @_;
	} elsif (ref $_[0] && ref $_[0] eq 'Gtk2::Gdk::Color') {
		return map { $_[0]->$_ >> 8 } qw(red green blue);
	} elsif (ref $_[0] && ref $_[0] eq 'ARRAY') {
		return @{$_[0]};
	} else {
		return &color_rgb(&color_gdk(@_));
	}
}

sub color_string {
	return sprintf('#%02x%02x%02x', &color_rgb(@_));
}

sub color_gdk {
	if (! defined $_[0]) {
		return Gtk2::Gdk::Color->new(0, 0, 0);
	} elsif (@_ == 3) {
		return Gtk2::Gdk::Color->new(
			map { $_ > 255 ? 65535 : $_ * 257 } @_);
	} elsif (ref $_[0] && ref $_[0] eq 'Gtk2::Gdk::Color') {
		return Gtk2::Gdk::Color->new(
			map { $_[0]->$_ } qw(red green blue));
	} elsif (ref $_[0] && ref $_[0] eq 'ARRAY') {
		return Gtk2::Gdk::Color->new(
			map { $_ > 255 ? 65535 : $_ * 257 } @{$_[0]});
	} else {
		return Gtk2::Gdk::Color->parse($_[0]);
	}
}

sub color_best_contrast {
	my $c = [ &color_rgb(shift) ];
	my @colors = @_ ? @_ : ([0, 0, 0], [255, 255, 255]);

	my $best = [ &color_rgb(shift @colors) ];
	my $best_dist = &color_distance($c, $best);
	while(@colors) {
		my $o = [ &color_rgb(shift @colors) ];
		my $o_dist = &color_distance($c, $o);
		($best, $best_dist) = ($o, $o_dist)
			if ($o_dist > $best_dist);
	}
	return $best;
}

sub color_distance {
	my @c1 = &color_rgb($_[0]);
	my @c2 = &color_rgb($_[1]);

	my @d = map { ($c1[$_] - $c2[$_]) ** 2 } (0..2);
	return $d[0] + $d[1] + $d[2];
}

1;

